/*
3d.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "3d.h"

#include "3d_curve.h"
#include "brush.h"
#include "bspspan.h"
#include "camera.h"
#include "color.h"
#include "display.h"
#include "edmodel.h"
#include "entity.h"
#include "error.h"
#include "geom.h"
#include "map.h"
#include "mdl.h"
#include "memory.h"
#include "message.h"
#include "poly.h"
#include "quest.h"
#include "qui.h"
#include "status.h"
#include "video.h"


static viewport_t *cur_vp;  // current viewport we're drawing to
static int cur_vport;
static int dx,dy;


void UpdateViewportBorder(int vport)
{
	int border_color;
   viewport_t *v;

	/* Draw viewport border */
	if (vport == M.display.active_vport)
		border_color = C_BORDER_ACTIVE;
	else
		border_color = C_BORDER_OTHER;

   v=&M.display.vport[vport];

   QUI_Box(v->xmin,v->ymin,v->xmax,v->ymax,border_color,border_color);
}


static void DrawBrushFaceEdges(brush_t *b,int face,int base_color,int force_bright)
{
	int   color;
	int   from, to;
	float  x0, y0, x1, y1;
	float  clipx, clipy;
	edge_t edge;
	int    found;
	int    was_fullbright;
	int    i, j;
	vec3_t *tverts;
	int   xs0, ys0;
	int   xs1, ys1;
	int   xs2, ys2;
	int   c0, c1;


	if (cur_vp->fullbright)
		was_fullbright = TRUE;
	else
		was_fullbright = FALSE;

	if (force_bright)
		cur_vp->fullbright = TRUE;

	tverts = b->tverts;
	for (i=0; i<b->num_edges;i++)
	{
		from = b->edges[i].startvertex;
		to = b->edges[i].endvertex;

		/* Check if edge is in face */
		found = FALSE;
		for (j=0; j<b->plane[face].num_verts; j++)
		{
			edge.startvertex = b->plane[face].verts[j];
			edge.endvertex = b->plane[face].verts[(j+1)%b->plane[face].num_verts];
			if (EdgesAreEqual(edge, b->edges[i]))
			{
				found = TRUE;
				break;
			}
		}
		if (!found)
			continue;

		switch (cur_vp->mode)
		{
			case NOPERSP:
				if (!cur_vp->fullbright)
				{
					if (((tverts[from].z < 0) && (tverts[to].z < 0)) ||
						 ((tverts[from].z > status.depth_clip) &&
						  (tverts[to].z > status.depth_clip)))
               {
					   continue;
               }
				}

				clipx = dx / cur_vp->zoom_amt;
				clipy = dy / cur_vp->zoom_amt;

				if (((tverts[from].x > clipx) && (tverts[to].x > clipx)) ||
				 	 ((tverts[from].x <-clipx) && (tverts[to].x <-clipx)) ||
					 ((tverts[from].y > clipy) && (tverts[to].y > clipy)) ||
					 ((tverts[from].y <-clipy) && (tverts[to].y <-clipy)))
				continue;

				x0 =  tverts[from].x * cur_vp->zoom_amt + cur_vp->xmin + dx;
				y0 = -tverts[from].y * cur_vp->zoom_amt + cur_vp->ymin + dy;
				x1 =  tverts[to  ].x * cur_vp->zoom_amt + cur_vp->xmin + dx;
				y1 = -tverts[to  ].y * cur_vp->zoom_amt + cur_vp->ymin + dy;

				xs0 = x0+0.5;
				ys0 = y0+0.5;
				xs1 = x1+0.5;
				ys1 = y1+0.5;

				restart_clip2d:

					c0 = OutCode2D(cur_vport, xs0, ys0);
					c1 = OutCode2D(cur_vport, xs1, ys1);

					if ((c0 | c1) == 0)
					{
						if (cur_vp->grid_type == ALIGN)
							Draw2DSpecialLine(cur_vport,xs0, ys0, xs1, ys1, base_color);
						else
							DrawLine(xs0, ys0, xs1, ys1, base_color);
						break;
					}
					else
					if ((c0 & c1) != 0) break;
					else
					if (c0 != 0)
					{
						Clip2D(cur_vport, xs0, ys0, xs1, ys1, &xs2, &ys2, c0);
						xs0 = xs2;
						ys0 = ys2;
					}
					else
					{
						Clip2D(cur_vport, xs0, ys0, xs1, ys1, &xs2, &ys2, c1);
						xs1 = xs2;
						ys1 = ys2;
					}
					goto restart_clip2d;
				break;

			case WIREFRAME:
         case SOLID:
				/* Immediately throw out lines outside view frustum */
                                
            if (OutsideFrustrum(tverts[from],tverts[to])) continue;

				if (!cur_vp->fullbright)
				{
					if ((tverts[from].z > 1) && (tverts[to].z > 1)) continue;
				}

				if ((!cur_vp->fullbright) && (!force_bright))
				{
					if (tverts[to].z < 0)
						color = base_color - tverts[from].z * 6;
					else
					if (b->tverts[from].z < 0)
						color = base_color - tverts[to].z * 6;
					else
					{
						color = (tverts[from].z + tverts[to].z) * 5;
						if (color > 15) color = 15;
						color = base_color - color;
					}
				}
				else
				   color = base_color;

				ClipDrawLine3D(cur_vport,
								   tverts[from].x,
								   tverts[from].y,
								   tverts[from].z,
								   tverts[to].x,
								   tverts[to].y,
								   tverts[to].z,
								   color);
				break;

		}
	}

	if (force_bright && !was_fullbright)
		cur_vp->fullbright = FALSE;
}

static void DrawBrushFace(brush_t *b,int face,int base_color,int force_bright)
{
   if (b->bt->flags&BR_F_EDGES)
   {
      DrawBrushFaceEdges(b,face,base_color,force_bright);
      return;
   }
}


void ReadPts(char *name)
{
   FILE *f;
   vec3_t t;

   if (M.npts)
   {
      Q_free(M.pts);
      Q_free(M.tpts);
   }

   f=fopen(name,"rt");
   if (!f)
      return;

   M.npts=0;
   M.pts=M.tpts=NULL;
   while (!feof(f))
   {
      if (fscanf(f,"%f %f %f",&t.x,&t.y,&t.z)!=3)
         break;
      M.pts=Q_realloc(M.pts,sizeof(vec3_t)*(M.npts+1));
      if (!M.pts)
      {
         fclose(f);
         HandleError("ReadPts","Out of memory!");
         M.npts=0;
         return;
      }
      M.pts[M.npts]=t;
      M.npts++;
   }
   fclose(f);

   M.tpts=Q_malloc(sizeof(vec3_t)*M.npts);
   if (!M.tpts)
   {
      Q_free(M.pts);
      HandleError("ReadPts","Out of memory!");
      M.npts=0;
      return;
   }

   NewMessage("%i points read",M.npts);
}

void FreePts(void)
{
   if (M.npts)
   {
      Q_free(M.pts);
      Q_free(M.tpts);
      M.npts=0;
      M.pts=M.tpts=NULL;
   }
}

static void DrawPts(matrix_t m)
{
   int i;

   int color;
   float x0,y0,x1,y1;
   float clipx,clipy;
   int xs0,xs1,xs2;
   int ys0,ys1,ys2;
   int c0,c1;

   svec_t s;

   if (!status.draw_pts)
      return;

   if (!M.npts)
      return;

   for (i=0;i<M.npts;i++)
   {
      Transform(M.tpts[i],M.pts[i],m);
   }

   for (i=0;i<M.npts-1;i++)
   {
      if (status.draw_pts==3)
      {
         color=CR_PTS_DIST;

         ToScreen(&s,M.tpts[i+1]);
         if (s.onscreen)
         {
            float dx,dy,dz;
            int x,y;

            dx=M.pts[i].x-M.pts[i+1].x;
            dy=M.pts[i].y-M.pts[i+1].y;
            dz=M.pts[i].z-M.pts[i+1].z;
            dx*=dx;
            dy*=dy;
            dz*=dz;
            dx=sqrt(dx+dy+dz);

            x=s.x-32;
            if (x<cur_vp->xmin+1) x=cur_vp->xmin+1;
            
            y=s.y-32;
            if (y<cur_vp->ymin+1) y=cur_vp->ymin+1;

            if (y+20<cur_vp->ymax)
               QUI_DrawStrM(x,y,cur_vp->xmax-2,-1,C_PTS_DIST_TEXT,0,0,"%0.1f",dx);
         }
      }
      else
      if (status.draw_pts==2)
      {
         if (i==M.npts-2)
            color=CR_PTS_BR_END;
         else
            color=CR_PTS_BR;

         ToScreen(&s,M.tpts[i]);
         if (s.onscreen) DrawVertex(cur_vport,s,CR_VERT);
      }
      else
      {
         if (i==(M.npts-2))
            color=CR_PTS_LEAK_END;
         else
         if (i==0)
            color=CR_PTS_LEAK_START;
         else
            color=CR_PTS_LEAK;
      }

		switch (cur_vp->mode)
		{
			case NOPERSP:
				if (!cur_vp->fullbright)
				{
					if (((M.tpts[i  ].z < 0) && (M.tpts[i+1].z < 0)) ||
						 ((M.tpts[i  ].z > status.depth_clip) &&
						  (M.tpts[i+1].z > status.depth_clip)))
					continue;
				}

				clipx = dx / cur_vp->zoom_amt;
				clipy = dy / cur_vp->zoom_amt;

				if (((M.tpts[i].x > clipx) && (M.tpts[i+1].x > clipx)) ||
					 ((M.tpts[i].x <-clipx) && (M.tpts[i+1].x <-clipx)) ||
					 ((M.tpts[i].y > clipy) && (M.tpts[i+1].y > clipy)) ||
					 ((M.tpts[i].y <-clipy) && (M.tpts[i+1].y <-clipy)))
				   continue;

				x0= M.tpts[i  ].x * cur_vp->zoom_amt + cur_vp->xmin + dx;
				y0=-M.tpts[i  ].y * cur_vp->zoom_amt + cur_vp->ymin + dy;
				x1= M.tpts[i+1].x * cur_vp->zoom_amt + cur_vp->xmin + dx;
				y1=-M.tpts[i+1].y * cur_vp->zoom_amt + cur_vp->ymin + dy;

				xs0 = x0+0.5;
				ys0 = y0+0.5;
				xs1 = x1+0.5;
				ys1 = y1+0.5;

				restart_clip2d:

					c0 = OutCode2D(cur_vport, xs0, ys0);
					c1 = OutCode2D(cur_vport, xs1, ys1);

					if ((c0 | c1) == 0)
					{
                  DrawLine(xs0, ys0, xs1, ys1, color);
						break;
					}
					else
					if ((c0 & c1) != 0) break;
					else
					if (c0 != 0)
					{
						Clip2D(cur_vport, xs0, ys0, xs1, ys1, &xs2, &ys2, c0);
						xs0 = xs2;
						ys0 = ys2;
					}
					else
					{
						Clip2D(cur_vport, xs0, ys0, xs1, ys1, &xs2, &ys2, c1);
						xs1 = xs2;
						ys1 = ys2;
					}
					goto restart_clip2d;
				break;

			case WIREFRAME:
         case SOLID:
				/* Immediately throw out lines outside view frustum */
        
            if (OutsideFrustrum(M.tpts[i],M.tpts[i+1])) continue;

				if (!cur_vp->fullbright)
				{
					if ((M.tpts[i].z > 1) && (M.tpts[i+1].z > 1)) continue;
				}

				ClipDrawLine3D(cur_vport,
								   M.tpts[i  ].x,
								   M.tpts[i  ].y,
                           M.tpts[i  ].z,
								   M.tpts[i+1].x,
								   M.tpts[i+1].y,
								   M.tpts[i+1].z,
								   color);
				break;

		}
   }
}

