/*
cmd/games/qengine/q_tex.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"


#include "qmap.h"

#include "game_reg.h"

#include "config.h"
#include "entclass.h"
#include "map.h"
#include "memory.h"
#include "s_error.h"
#include "s_texture.h"
#include "s_types.h"
#include "status.h"
#include "texcat.h"
#include "token.h"


static config_var_t
   *cv_wad_path,
   *cv_entity_path;


typedef struct
{
	char  name[16];      /* Name of the texture */
	s_u32 width;         /* Width of texture, multiple of 8 */
	s_u32 height;        /* Height of texture, multiple of 8 */
	s_u32 offset1;       /* Offset to 1 x 1 data */
	s_u32 offset2;       /* Offset to 1/2 x 1/2 data */
	s_u32 offset4;       /* Offset to 1/4 x 1/4 data */
	s_u32 offset8;       /* Offset to 1/8 x 1/8 data */
} miptex_t;

typedef struct
{
   char magic[4];       /* "WAD2", Name of the new WAD format */
   s_u32 numentries;    /* Number of entries */
   s_u32 diroffset;     /* Position of WAD directory in file */
} wadhead_t;

typedef struct
{
   s_u32 offset;        /* Position of the entry in WAD */
   s_u32 dsize;         /* Size of the entry in WAD file */
   s_u32 size;          /* Size of the entry in memory */
   s_u8  type;          /* type of entry */
   s_u8  cmprs;         /* Compression. 0 if none. */
   s_u16 dummy;         /* Not used */
   char  name[16];      /* 1 to 16 characters, '\0'-padded */
} wadentry_t;

/*
Types:
 0x40 = Color Palette
 0x42 = Pictures for status bar
 0x44 = Mip Texture
 0x45 = Console picture (flat)
*/


typedef struct
{
   int ofs;
} g_t;

static FILE *wad;


static void LoadTex(void *tex,const texture_t *t)
{
   unsigned char *buf;
   miptex_t miptex;
   FILE *f=wad;

   if (!wad) return;

   fseek(f,((g_t *)t->g)->ofs,SEEK_SET);
   fread(&miptex,sizeof(miptex_t),1,f);

   S_LittleEndian(miptex.width);
   S_LittleEndian(miptex.height);
   S_LittleEndian(miptex.offset1);
   S_LittleEndian(miptex.offset2);
   S_LittleEndian(miptex.offset4);
   S_LittleEndian(miptex.offset8);

   fseek(f,((g_t *)t->g)->ofs+miptex.offset1,SEEK_SET);
   buf=Q_malloc(miptex.width*miptex.height);
   if (!buf)
      return;
   fread(buf,miptex.width*miptex.height,1,f);
   S_T_AddData8(tex,miptex.width,miptex.height,0,0,buf,NULL);
   Q_free(buf);
}

static tex_funcs_t quake_tex={LoadTex};

static int LoadWad(const char *wadname,int verbose)
{
/* the currently loaded .wad file */
static char cur_wad[MAX_PATH];

   FILE *f;
   int i;
   int found_pal;

   wadhead_t head;
   wadentry_t *entry;

   unsigned char tex_pal[768];
   

   if (!strcmp(cur_wad,wadname))
      return 1;

   S_T_Clear();
   S_T_Init(0,&quake_tex);

   strcpy(cur_wad,wadname);

   if (wad)
   {
      fclose(wad);
      wad=NULL;
   }
   
   wad = f = fopen(wadname,"rb");
   if (!f)
   {
      if (verbose)
         S_Error("LoadWad","Can't open '%s'!",wadname);
      return 0;
   }
   
   fread(&head,sizeof(wadhead_t),1,f);

   if ((head.magic[0]!='W') ||
       (head.magic[1]!='A') ||
       (head.magic[2]!='D') ||
       (head.magic[3]!='2'))
   {
      fclose(f);
      wad=NULL;
      return 0;
   }
   S_LittleEndian(head.diroffset);
   S_LittleEndian(head.numentries);
   
   fseek(f,head.diroffset,SEEK_SET);
   
	entry = (wadentry_t *)Q_malloc(sizeof(wadentry_t) * head.numentries);
   if (!entry)
   {
      if (verbose)
         S_Error("LoadWad","Out of memory!");
      fclose(f);
      wad=NULL;
      return 0;
   }
	fread(entry, sizeof(wadentry_t) * head.numentries, 1, f);

   found_pal=0;
   for (i=0;i<head.numentries;i++)
   {
      S_LittleEndian(entry[i].offset);
      S_LittleEndian(entry[i].dsise);
      S_LittleEndian(entry[i].size);
   
      switch (entry[i].type)
      {
      case 0x40:
			fseek(f,entry[i].offset,SEEK_SET);
			fread(tex_pal, sizeof(tex_pal), 1, f);
         S_T_SetPalette(tex_pal);
         found_pal=1;
         break;
      }
   }

   if (!found_pal)
   { /* TODO */
#if 0
      FILE *f2;

      f=S_OpenFile("quake.pal","rb");
      if (!f2)
      {
         S_Error("LoadWad","No palette in .wad file, and couldn't open quake.pal");
      }
      else
      {
         fread(tex_pal,sizeof(tex_pal),1,f2);
         S_T_SetPalette(tex_pal);
         fclose(f2);
      }
#endif
   }

   for (i=0;i<head.numentries;i++)
   {
      if (entry[i].type!=0x44) continue;

      {
         g_t *g;
         struct {s_u32 x; s_u32 y; } h;

         g=Q_malloc(sizeof(g_t));
         if (!g) return 0;
         g->ofs=entry[i].offset;
         strlwr(entry[i].name);
         
         fseek(f,entry[i].offset+16,SEEK_SET);
         fread(&h,sizeof(h),1,f);
   
         S_LittleEndian(h.x);
         S_LittleEndian(h.y);
         
         if (!S_T_AddTexture(entry[i].name,h.x,h.y,g))
            return 0;
      }
   }

	Q_free(entry);

   return 1;
}

/* TODO
static int Q_TexBSPFlags(texdef_t *t)
{
   if (t->name[0]=='*')             return TEX_FULLBRIGHT|TEX_NONSOLID;
   if (!strnicmp(t->name,"SKY",3))  return TEX_FULLBRIGHT;
   if (!stricmp(t->name,"TRIGGER")) return TEX_NODRAW;
   if (!stricmp(t->name,"CLIP"))    return TEX_NODRAW|TEX_NONSOLID;
   return 0;
}
*/

typedef struct G_texdef_s
{
   int detail;
} G_texdef_t;


static int QT_Init(texdef_t *t)
{
   if (!t->g)
   {
      t->g=Q_malloc(sizeof(G_texdef_t));
      if (!t->g) return 0;
   }
   t->g->detail=0;
   return 1;
}

static void QT_Free(texdef_t *t)
{
   if (t->g)
   {
      Q_free(t->g);
      t->g=NULL;
   }
}

static int QT_Copy(texdef_t *dst,const texdef_t *src)
{
   if (!src->g) return 1;
   if (!dst->g)
   {
      dst->g=Q_malloc(sizeof(G_texdef_t));
      if (!dst->g) return 0;
   }
   dst->g->detail=src->g->detail;
   return 1;
}


static game_reg_t G_Quake=
{
   "Quake",
   {
      QMap_Load,
      QMap_Save
   },
   {
      QT_Init,
      QT_Free,
      QT_Copy
   },
};


static int Q_LoadTexInfo(texdef_t *t)
{
   if (!t->g)
   {
      t->g=Q_malloc(sizeof(G_texdef_t));
      if (!t->g) return ERROR_NOMEM;
   }
   t->g->detail=0;

   while (TokenAvailable(0))
   {
      GETTOKEN(0,TOKEN_NAME);
      if (!strcmp(token,"detail"))
         t->g->detail=1;
      else
         ; /* TODO */
/*         return ERROR_PARSE;*/
   }

   return ERROR_NO;
}

static void Q_SaveTexInfo(texdef_t *t,FILE *fp)
{
   if (t->g && t->g->detail)
      fprintf(fp," detail");
}


#include "cmds.h"

#include "entity.h"
#include "hook.h"
#include "quest.h"
#include "s_gui.h"
#include "s_message.h"
#include "s_sys.h"
#include "uid.h"

static void C_select_wad(void)
{
   char filename[MAX_PATH];

   filename[0]=0;
   if (!S_GUI_FileDialog("Select .wad","Load","*.wad",filename))
      return;

   SetKeyValue(UID_Entity(M.WorldSpawn),"wad",filename);
   LoadWad(filename,1);
   Map_Update();
   S_NewMessage("Wadfile set to: %s",filename);
}

static cmd_reg_t quake1_cmds[]={
{"quake1.select_wad",C_select_wad,{"Misc/Quake 1/Select wad..."},{},"Select a .wad file to use for textures."}
};


static void q1_map_new(void)
{
   entity_t *e=UID_Entity(M.WorldSpawn);
   if (e)
      SetKeyValue(e,"wad",cv_wad_path->s);
   LoadWad(cv_wad_path->s,1);
}

static void q1_map_load(void)
{
   entity_t *e=UID_Entity(M.WorldSpawn);
   const char *c;
   if (!e) return;
   c=GetKeyValue(e,"wad");
   if (!c || !Sys_Exists(c)) return;
   LoadWad(c,1);
}


#include "plugin.h"
PLUGIN_HEAD

static int Done(void);

static plugin_t plugin=
{{Init,Done},{"games/quake","Game plugin for Quake I and games/mods based on it (e.g. Hexen 2)."}};

static int Init(void)
{ /* TODO */
   cv_wad_path=Cfg_DefVar("quake1.wad_path","/mnt/msdos_d/quake/id1/tex.wad","Path to the default .wad file to load textures from.");
   cv_entity_path=Cfg_DefVar("quake1.entity_path",
      "games/quake/es/*",
      "Path to the entity set to use for Quake 1.");

   Cfg_SetInt(cv_texture_align_rotate,0);
   Cfg_SetInt(cv_texture_align_shiftx,0);
   Cfg_SetInt(cv_texture_align_shifty,0);
   Cfg_SetFloat(cv_texture_align_scalex,1);
   Cfg_SetFloat(cv_texture_align_scaley,1);

/* TODO: check for other game before unloading map */
   if (!Map_Close("Change game"))
      return 0;
   Map_Unload();
      
   if (!G_Load(&G_Quake))
   {
      S_Error("games/quake","Can't load, another game is already loaded!");
      return 0;
   }

   QMap_Init();
   
   qmap_loadtexinfo=Q_LoadTexInfo;
   qmap_savetexinfo=Q_SaveTexInfo;
   
   Cmd_Register(&plugin,quake1_cmds,sizeof(quake1_cmds));

   Hook_AddFunc(hk_map_new,HOOK_FUNC_VOID,(hook_func_void_t)q1_map_new,NULL,&plugin);
   Hook_AddFunc(hk_map_load,HOOK_FUNC_VOID,(hook_func_void_t)q1_map_load,NULL,&plugin);
   
   InitEClass(cv_entity_path->s);

   TC_Init();
   TC_Load("games/quake/texcat.txt");
   
   Map_New();
   return 1;
}

static int Done(void)
{
   if (!Map_Close("Unload Quake 1"))
      return 0;
   Map_Unload();
   G_Unload();
   S_T_Clear();
   TC_Done();
   Map_New();
   return 1;
}

PLUGIN_FOOT

