#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "config.h"

#include "s_error.h"
#include "token.h"
#include "version.h"


config_var_t **cvars;
int num_cvars;


static config_var_t *Cfg_GetVar(const char *name)
{ /* TODO: insertion sort, would make search quicker */
  int i;
  config_var_t *c;

  for (i=0;i<num_cvars;i++)
     if (!strcmp(name,cvars[i]->name))
        return cvars[i];

  c=malloc(sizeof(config_var_t));
  if (!c) S_Abort("Cfg_GetVar","Out of memory!");
  memset(c,0,sizeof(config_var_t));
  c->name=strdup(name);
  if (!c->name) S_Abort("Cfg_GetVar","Out of memory!");

  cvars=realloc(cvars,sizeof(config_var_t *)*(num_cvars+1));
  if (!cvars) S_Abort("Cfg_GetVar","Out of memory!");
  cvars[num_cvars++]=c;
  return c;
}


config_var_t *Cfg_DefVar(const char *name,const char *value,const char *desc)
{
   config_var_t *c;

   c=Cfg_GetVar(name);
   if (!c->desc && desc)
   {
      c->desc=strdup(desc);
      if (!c->desc) S_Abort("Cfg_DefVar","Out of memory!");
   }
   if (!c->s && value)
      Cfg_SetStr(c,value);
   return c;
}


int Cfg_Int(const char *name)
{
   config_var_t *c;
   c=Cfg_GetVar(name);
   return c->i;
}

float Cfg_Float(const char *name)
{
   config_var_t *c;
   c=Cfg_GetVar(name);
   return c->f;
}

const char *Cfg_Str(const char *name)
{
   config_var_t *c;
   c=Cfg_GetVar(name);
   return c->s;
}


void Cfg_SetInt(config_var_t *c,int i)
{
   char buf[128];
   
   c->i=i;
   c->f=i;
   snprintf(buf,sizeof(buf)-1,"%i",i);
   if (c->s) free(c->s);
   c->s=strdup(buf);
   if (!c->s) S_Abort("Cfg_SetInt","Out of memory!");
}

void Cfg_SetFloat(config_var_t *c,float f)
{
   char buf[128];
   
   c->i=f;
   c->f=f;
   snprintf(buf,sizeof(buf)-1,"%g",f);
   if (c->s) free(c->s);
   c->s=strdup(buf);
   if (!c->s) S_Abort("Cfg_SetFloat","Out of memory!");
}

void Cfg_SetStr(config_var_t *c,const char *str)
{
   char *e;

   if (c->s) free(c->s);
   if (str)
   {
      c->s=strdup(str);
      if (!c->s) S_Abort("Cfg_SetStr","Out of memory!");
   
      c->f=strtod(str,&e);
      if (*e) c->f=0;
      c->i=strtol(str,&e,0);
      if (*e) c->i=0;
   }
   else
   {
      c->s=NULL;
      c->f=0;
      c->i=0;
   }
}


void Config_Load(const char *fname)
{
   char buf[256];
   
   if (!TokenFile(fname,TOKEN_C|TOKEN_NUMBER|TOKEN_NAME|TOKEN_MISC|TOKEN_STRING,NULL))
      return;

   while (TokenGet(1,-1))
   {
      if (strlen(token)>sizeof(buf)-1)
         S_Abort("Config_Load","Key too long at line %i, %s.",
            token_linenum,fname);
      strcpy(buf,token);
      if (!TokenGet(1,-1))
         S_Abort("Config_Load","Parse error on line %i, %s.",
            token_linenum,fname);
            
      while (strcmp(token,"="))
      {
         if (strlen(buf)+strlen(token)>sizeof(buf)-1)
            S_Abort("Config_Load","Key too long at line %i, %s.",
               token_linenum,fname);
         strcat(buf,token);
         if (!TokenGet(1,-1))
            S_Abort("Config_Load","Parse error on line %i, %s.",
               token_linenum,fname);
      }
      if (!TokenGet(1,-1))
         S_Abort("Config_Load","Parse error on line %i, %s.",
            token_linenum,fname);
      if (token[0]=='"')
      {
         token[strlen(token)-1]=0;
         Cfg_SetStr(Cfg_GetVar(buf),&token[1]);
      }
      else
         Cfg_SetStr(Cfg_GetVar(buf),token);

      if (!TokenGet(1,-1))
         S_Abort("Config_Load","Parse error on line %i, %s.",
            token_linenum,fname);
      if (strcmp(token,";"))
         S_Abort("Config_Load","Parse error on line %i, %s.",
            token_linenum,fname);
   }
}

/* TODO: keep a quest-save marker in file, don't change anything before it,
or any variables defined there, only rewrite section after it.
*/
void Config_Save(const char *fname)
{
   FILE *f;
   int i;
   config_var_t **c;

   f=fopen(fname,"wt");
   if (!f)
      return;

   fprintf(f,
      "/*\n"
      "Configuration file for Quest %s\n"
      "*/\n\n",
      QUEST_VERSION);

   for (c=cvars,i=num_cvars;i;i--,c++)
   {
      fprintf(f,"\n");
      if ((*c)->desc)
         fprintf(f,"/* %s */\n",(*c)->desc);
      fprintf(f,"%s = \"%s\";\n",
         (*c)->name,(*c)->s?(*c)->s:"");
   }

   fclose(f);
}

