#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <GL/gl.h>

#include "defines.h"
#include "types.h"

#include "gl_rend.h"
#include "s_render.h"
#include "s_trans.h"

#include "s_texture.h"
#include "gl_texture.h"

#include "brush.h"
#include "camera.h"
#include "config.h"
#include "entclass.h"
#include "entity.h"
#include "geom.h"
#include "matrix.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "status.h"


/*
TODO: check if curves are rendered correctly here. The glEvalMesh stuff
doesn't match v2.4, check which one matches Quake 3.
*/


/*
This file provides OpenGL based implementations of the functions in gl_rend.h,
s_render.h, and s_trans.h, except S_R_GetMousePos(), which some system
dependant code will have to handle somewhere.
*/


#define Z_NEAR 1.0
#define Z_FAR  2048.0


static config_var_t
   *cv_gl_curve_subdivide,
   *cv_gl_use_display_lists;


/*
Workaround for buggy nvidia OpenGL driver that can't correctly handle
multiple glVertex:s in one glBegin(GL_POINTS)/glEnd() pair.
*/
#define BUGGY_NVIDIA_OPENGL

#ifdef BUGGY_NVIDIA_OPENGL

#define P_glBegin(x)
#define P_glEnd()
#define P_glVertex3fv(x) do { glBegin(GL_POINTS); glVertex3fv(x); glEnd(); } while (0)

#else

#define P_glBegin(x) glBegin(GL_POINTS)
#define P_glEnd() glEnd()
#define P_glVertex3fv(x) glVertex3fv(x)

#endif


#define FOR_EACH_SEL  for (s=&M.sel,i=-1;i<M.num_rsel;s=(++i<M.num_rsel?M.rsel[i]:NULL))


typedef struct
{
   float c[4];
   const char *name,*desc;
} render_color_t;

static render_color_t colors[]=
{
/* Used before texture-mapping */
#define C_BLANK               (colors[ 0].c)
{{1.0, 1.0, 1.0, 1.0},
   "blank",
   "Color used when texturing. You probably don't want to change this."},

#define C_BACK                (colors[ 1].c)
{{0.0, 0.0, 0.0, 0.0},"back","Background."},

#define C_GRID_SMALL          (colors[ 2].c)
{{0.0, 0.0, 0.7, 1.0},"grid_small","Small grid."},
#define C_GRID_LARGE          (colors[ 3].c)
{{0.1, 0.1, 1.0, 1.0},"grid_large","Large grid."},

#define C_BRUSH_NORMAL        (colors[ 4].c)
{{1.0, 1.0, 1.0, 1.0},"brush_normal","Normal brushes (wireframe)."},
#define C_BSEL                (colors[ 5].c)
{{1.0, 1.0, 0.0, 1.0},"bsel","Selected brushes (wireframe)."},
#define C_BSEL_TEX            (colors[ 6].c)
{{1.0, 1.0, 0.2, 0.2},"bsel_tex","Selected brushes (textured)."},
#define C_BSEL_TEX_2          (colors[ 7].c)
{{1.0, 0.5, 0.5, 1.0},"bsel_tex_2","Wireframe version of selected brushes in textured mode."},
#define C_BRUSH_FOCAL         (colors[ 8].c)
{{1.0, 0.0, 0.0, 1.0},"brush_focal","Normal brush focal points."},
#define C_BSEL_FOCAL          (colors[ 9].c)
{{1.0, 0.0, 0.0, 1.0},"bsel_focal","Focal point of selected brushes."},

#define C_VERT                (colors[10].c)
{{0.0, 1.0, 1.0, 1.0},"vert","Vertices (wireframe)."},
#define C_VERT_TEX            (colors[11].c)
{{0.0, 1.0, 1.0, 1.0},"vert_tex","Vertices (textured)."},
#define C_VSEL                (colors[12].c)
{{1.0, 0.0, 0.0, 1.0},"vsel","Selected vertices."},

#define C_FACE_FOCAL          (colors[13].c)
{{0.1, 1.0, 0.1, 1.0},"face_focal","Face focal points."},
#define C_FSEL                (colors[14].c)
{{1.0, 1.0, 0.0, 1.0},"fsel","Selected faces (wireframe)."},
#define C_FSEL_TEX            (colors[15].c)
{{1.0, 1.0, 0.2, 0.2},"fsel_tex","Selected faces (textured)."},
#define C_FSEL_TEX_2          (colors[16].c)
{{1.0, 0.5, 0.5, 1.0},"fsel_tex_2","Wireframe version of selected faces in textured mode."},

#define C_ENTITY_FOCAL        (colors[17].c)
{{0.8, 0.8, 0.8, 1.0},"entity_focal","Entity focal points."},
#define C_ESEL                (colors[18].c)
{{1.0, 1.0, 0.0, 1.0},"esel","Selected entities."},

#define C_SELECTION_BOX       (colors[19].c)
{{0.0, 0.0, 1.0, 1.0},"selection_box","Selection box."},

#define C_LINK_TARGET_TARGETNAME (colors[20].c)
{{0.0, 1.0, 1.0, 1.0},"link_target_targetname","target/targetname link arrows."},

#define C_LINK_PATHTARGET_TARGETNAME (colors[21].c)
{{1.0, 0.5, 0.5, 1.0},"link_pathtarget_targetname","pathtarget/targetname link arrows."},

#define C_LINK_KILLTARGET_TARGETNAME (colors[22].c)
{{1.0, 0.8, 0.0, 1.0},"link_killtarget_targetname","killtarget/targetname link arrows."},

#define C_MAX 23

};


typedef struct S_r_face_s
{
   gl_texture_t *t;
   float light;
} r_face_t;


static void DrawLink(entity_t *e,const char *key,const char *dst_name)
{
   entity_t *e2;
   vec3_t delta,vr,vu,t,t2;
   const char *d;

   for (e2=M.VEntityHead;e2;e2=e2->VNext)
   {
      d=GetKeyValue(e2,dst_name);
      if (!d) continue;
      if (strcmp(d,key)) continue;
      glBegin(GL_LINES);
      glVertex3fv((float *)&e->center);
      glVertex3fv((float *)&e2->center);
      glEnd();

      VEC_Sub(delta,e2->center,e->center);
      Normalize(&delta);

      vr.x=vr.y=vr.z=0;
      if (fabs(delta.x)>fabs(delta.y))
         vr.y=1;
      else
         vr.x=1;
      CrossProd(vr,delta,&vu);
      Normalize(&vu);
      CrossProd(vu,delta,&vr);
      vr.x*=8; vr.y*=8; vr.z*=8;
      vu.x*=8; vu.y*=8; vu.z*=8;
      delta.x*=32; delta.y*=32; delta.z*=32;

      glBegin(GL_TRIANGLE_FAN);
      glVertex3fv((float *)&e2->center);

      VEC_Sub(t2,e2->center,delta);
      VEC_Sub(t2,t2,vr);
      VEC_Sub(t2,t2,vu);
      glVertex3fv((float *)&t2);
      
      VEC_Sub(t,e2->center,delta);
      VEC_Sub(t,t,vr);
      VEC_Add(t,t,vu);
      glVertex3fv((float *)&t);
      
      VEC_Sub(t,e2->center,delta);
      VEC_Add(t,t,vr);
      VEC_Add(t,t,vu);
      glVertex3fv((float *)&t);
      
      VEC_Sub(t,e2->center,delta);
      VEC_Add(t,t,vr);
      VEC_Sub(t,t,vu);
      glVertex3fv((float *)&t);
      
      glVertex3fv((float *)&t2);
      
      glEnd();
   }
}

static void DrawLinks(void)
{
   entity_t *e;
   const char *c;

   glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
   for (e=M.VEntityHead;e;e=e->VNext)
   {
      if ((c=GetKeyValue(e,"target")))
      {
         glColor4fv(C_LINK_TARGET_TARGETNAME);
         DrawLink(e,c,"targetname");
      }
      if ((c=GetKeyValue(e,"pathtarget")))
      {
         glColor4fv(C_LINK_PATHTARGET_TARGETNAME);
         DrawLink(e,c,"targetname");
      }
      if ((c=GetKeyValue(e,"killtarget")))
      {
         glColor4fv(C_LINK_KILLTARGET_TARGETNAME);
         DrawLink(e,c,"targetname");
      }
   }
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}


static void Wire_Brush(brush_t *b)
{
   int i;
   edge_t *e;

   if (b->bt->type==BRUSH_Q3_CURVE)
   {
      glEnd();
      glMapGrid2f(
         cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,0,1,
         cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2,0,1);
      glEnable(GL_MAP2_VERTEX_3);
      glMap2f(GL_MAP2_VERTEX_3,0,1,3*b->x.q3c.sizex,b->x.q3c.sizey,0,1,3,b->x.q3c.sizex,(float *)b->verts);
      glEvalMesh2(GL_LINE,
         0,cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,
         0,cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2);
      glDisable(GL_MAP2_VERTEX_3);
      glBegin(GL_LINES);
      return;
   }

   for (i=0,e=b->edges;i<b->num_edges;i++,e++)
   {
      glVertex3fv((float *)&b->verts[e->startvertex]);
      glVertex3fv((float *)&b->verts[e->endvertex]);
   }
}

static void Wire_Ent(entity_t *e)
{
   glBegin(GL_LINE_LOOP);
   glVertex3f(e->min.x,e->min.y,e->min.z);
   glVertex3f(e->min.x,e->max.y,e->min.z);
   glVertex3f(e->max.x,e->max.y,e->min.z);
   glVertex3f(e->max.x,e->min.y,e->min.z);
   glEnd();
   glBegin(GL_LINE_LOOP);
   glVertex3f(e->min.x,e->min.y,e->max.z);
   glVertex3f(e->min.x,e->max.y,e->max.z);
   glVertex3f(e->max.x,e->max.y,e->max.z);
   glVertex3f(e->max.x,e->min.y,e->max.z);
   glEnd();
   glBegin(GL_LINES);
   glVertex3f(e->min.x,e->min.y,e->min.z);
   glVertex3f(e->min.x,e->min.y,e->max.z);
   glVertex3f(e->max.x,e->min.y,e->min.z);
   glVertex3f(e->max.x,e->min.y,e->max.z);
   glVertex3f(e->min.x,e->max.y,e->min.z);
   glVertex3f(e->min.x,e->max.y,e->max.z);
   glVertex3f(e->max.x,e->max.y,e->min.z);
   glVertex3f(e->max.x,e->max.y,e->max.z);
   glEnd();
}

static void Wire_Draw(void)
{
   brush_t *b;
   bsel_t *bs;
   
   vec3_t *v;
   vsel_t *vs;

   sel_t *s;
   
   int i;
   

   if (cv_render_flags->i&RENDER_BRUSH)
   {
      glColor4fv(C_BRUSH_NORMAL);
      glBegin(GL_LINES);
      for (b=M.VBrushHead;b;b=b->VNext)
         if (b->drawn==1)
            Wire_Brush(b);

      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=SEL_DRAW)
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
            Wire_Brush(bs->Brush);
      }
      glEnd();

      glColor4fv(C_BSEL);
      glBegin(GL_LINES);

      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
            Wire_Brush(bs->Brush);
      }

      glEnd();


      P_glBegin(GL_POINTS);

      glColor4fv(C_BSEL_FOCAL);

      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
         {
            b=bs->Brush;
            P_glVertex3fv((float *)(&b->center));
         }
      }

      if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
      {
         glColor4fv(C_BRUSH_FOCAL);
         for (b=M.VBrushHead;b;b=b->VNext)
         {
            if (b->drawn!=1) continue;
            P_glVertex3fv((float *)(&b->center));
         }

         FOR_EACH_SEL
         {
            if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=SEL_DRAW)
               continue;
            for (bs=s->bsel;bs;bs=bs->Next)
            {
               b=bs->Brush;
               if (b->drawn!=1) continue;
               P_glVertex3fv((float *)(&b->center));
            }
         }

         if (cv_render_flags->i&RENDER_BRUSH_VERT)
         {
            glColor4fv(C_VERT);
            FOR_EACH_SEL
            {
               if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
                  continue;
               for (bs=s->bsel;bs;bs=bs->Next)
               {
                  b=bs->Brush;
                  for (i=0,v=b->verts;i<b->num_verts;i++,v++)
                  {
                     P_glVertex3fv((float *)v);
                  }
               }
            }
         }
      }
      P_glEnd();
   }

   if (cv_render_flags->i&RENDER_FACE_FOCAL)
   {
      face_t *f;
      fsel_t *fs;
      int j;
      
      glColor4fv(C_FACE_FOCAL);
      P_glBegin(GL_POINTS);
      for (b=M.VBrushHead;b;b=b->VNext)
      {
         for (f=b->faces,i=b->num_faces;i;i--,f++)
         {
            P_glVertex3fv((float *)&f->center);
         }
      }
      FOR_EACH_SEL
      {
         if (!(s->flags&SEL_DRAW))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
         {
            b=bs->Brush;
            for (f=b->faces,i=b->num_faces;i;i--,f++)
            {
               P_glVertex3fv((float *)&f->center);
            }
         }
      }
      P_glEnd();
      
      glColor4fv(C_FSEL);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
                  
         for (fs=s->fsel;fs;fs=fs->Next)
         {
            b=fs->Brush;
            f=&b->faces[fs->num];
            glBegin(GL_LINE_LOOP);
            for (j=0;j<f->num_verts;j++)
               glVertex3fv((float *)&(b->verts[f->verts[j]]));
            glEnd();
         }
      }

      if (cv_render_flags->i&RENDER_FACE_VERT)
      {
         glColor4fv(C_VERT);
         P_glBegin(GL_POINTS);
         FOR_EACH_SEL
         {
            if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
               continue;
            for (fs=s->fsel;fs;fs=fs->Next)
            {
               b=fs->Brush;
               f=&b->faces[fs->num];
               for (j=0;j<f->num_verts;j++)
               {
                  P_glVertex3fv((float *)&b->verts[f->verts[j]]);
               }
            }
         }
         P_glEnd();
      }
   }
   
   glColor4fv(C_VSEL);
   P_glBegin(GL_POINTS);
   FOR_EACH_SEL
   {
      if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
         continue;
      for (vs=s->vsel;vs;vs=vs->Next)
      {
         P_glVertex3fv((float *)&(vs->Brush->verts[vs->num]));
      }
   }
   P_glEnd();

   if (cv_render_flags->i&RENDER_ENTITY)
   {
      entity_t *e;
      esel_t *es;

      for (e=M.VEntityHead;e;e=e->VNext)
         if (e->drawn==1 && e->uid!=M.WorldSpawn)
         {
            glColor4f(
               e->eclass->color[0],
               e->eclass->color[1],
               e->eclass->color[2],
               1.0);
            Wire_Ent(e);
         }
      FOR_EACH_SEL
      {
         if ((s->flags&SEL_DRAW)!=SEL_DRAW)
            continue;
         for (es=s->esel;es;es=es->Next)
            Wire_Ent(es->Entity);
      }

      if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
      {
         glColor4fv(C_ENTITY_FOCAL);
         P_glBegin(GL_POINTS);
         for (e=M.VEntityHead;e;e=e->VNext)
         {
            P_glVertex3fv((float *)&(e->center));
         }
         FOR_EACH_SEL
         {
            if ((s->flags&SEL_DRAW)!=SEL_DRAW)
               continue;
            for (es=s->esel;es;es=es->Next)
            {
               P_glVertex3fv((float *)&(es->Entity->center));
            }
         }
         P_glEnd();
      }

      glColor4fv(C_ESEL);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (es=s->esel;es;es=es->Next)
            Wire_Ent(es->Entity);
      }
   }

   if (cv_render_flags->i&RENDER_ENTITY_LINK)
      DrawLinks();
}


static void usetex(gl_texture_t *t)
{
static int cur_tex=0;
   if (!t)
   {
      if (cur_tex!=0)
      {
         glDisable(GL_TEXTURE_2D);
         glBindTexture(GL_TEXTURE_2D,0);
         cur_tex=0;
      }
   }
   else
   {
      if (cur_tex==0)
      {
         glEnable(GL_TEXTURE_2D);
         glColor4fv(C_BLANK);
      }
      if (cur_tex!=t->num)
      {
         glBindTexture(GL_TEXTURE_2D,t->num);
         cur_tex=t->num;
      }
   }
}

static void Flat_Brush(brush_t *b)
{
   int i,j;
   face_t *f;
   vec3_t *v;

   if (b->bt->type==BRUSH_Q3_CURVE)
   { /* TODO */
      glColor4f(0.8,0.8,0.8,1.0); /* TODO */
      glMapGrid2f(
         cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,0,1,
         cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2,0,1);
      glEnable(GL_MAP2_VERTEX_3);
      glMap2f(GL_MAP2_VERTEX_3,
         0,1,3*b->x.q3c.sizex,b->x.q3c.sizey,
         0,1,3,b->x.q3c.sizex,
         (float *)b->verts);
      glEvalMesh2(GL_FILL,
         0,cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,
         0,cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2);
      glDisable(GL_MAP2_VERTEX_3);
      return;
   }
   
   for (f=b->faces,i=b->num_faces;i;i--,f++)
   {
      glColor4f(f->r->light,f->r->light,f->r->light,1);
      glBegin(GL_POLYGON);
      for (j=0;j<f->num_verts;j++)
      {
         v=&b->verts[f->verts[j]];
         glVertex3f(v->x,v->y,v->z);
      }
      glEnd();
   }
}

static void Tex_Brush(brush_t *b)
{
   int i,j;
   face_t *f;
   vec3_t *v;
   gl_texture_t *t;

   if (b->bt->type==BRUSH_Q3_CURVE)
   { /* TODO */
      t=S_GL_LoadTexture(b->tex.name);
      usetex(t);
      if (t)
      {
         glMapGrid2f(
            cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,0,1,
            cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2,0,1);
         glEnable(GL_MAP2_VERTEX_3);
         glEnable(GL_MAP2_TEXTURE_COORD_2);

         glMap2f(GL_MAP2_VERTEX_3,
            0,1,3*b->x.q3c.sizex,b->x.q3c.sizey,
            0,1,3,b->x.q3c.sizex,
            (float *)b->verts);
         glMap2f(GL_MAP2_TEXTURE_COORD_2,
            0,1,2*b->x.q3c.sizex,b->x.q3c.sizey,
            0,1,2,b->x.q3c.sizex,
            (float *)b->x.q3c.a);
            
         glEvalMesh2(GL_FILL,
            0,cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,
            0,cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2);
         
         glDisable(GL_MAP2_TEXTURE_COORD_2);
         glDisable(GL_MAP2_VERTEX_3);
      }
      else
      {
         glColor4f(0.8,0.8,0.8,1.0); /* TODO */
         glMapGrid2f(
            cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,0,1,
            cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2,0,1);
         glEnable(GL_MAP2_VERTEX_3);
         glMap2f(GL_MAP2_VERTEX_3,
            0,1,3*b->x.q3c.sizex,b->x.q3c.sizey,
            0,1,3,b->x.q3c.sizex,
            (float *)b->verts);
         glEvalMesh2(GL_FILL,
            0,cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,
            0,cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2);
         glDisable(GL_MAP2_VERTEX_3);
      }
      return;
   }
   
   for (f=b->faces,i=b->num_faces;i;i--,f++)
   {
      t=f->r->t;
      usetex(t);
      if (t)
      {
         glColor4fv(C_BLANK);
         glBegin(GL_POLYGON);
         for (j=0;j<f->num_verts;j++)
         {
            v=&b->verts[f->verts[j]];
            glTexCoord2f(f->texv[j].x/t->tex.width,f->texv[j].y/t->tex.height);
            glVertex3f(v->x,v->y,v->z);
         }
         glEnd();
      }
      else
      {
         glColor4f(f->r->light,f->r->light,f->r->light,1);
         glBegin(GL_POLYGON);
         for (j=0;j<f->num_verts;j++)
         {
            v=&b->verts[f->verts[j]];
            glVertex3f(v->x,v->y,v->z);
         }
         glEnd();
      }
   }
}

static void Flat_BSel(brush_t *b)
{
   face_t *f;
   int i,j;
   vec3_t *v;
   
   if (b->bt->type==BRUSH_Q3_CURVE)
   {
      glMapGrid2f(
         cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,0,1,
         cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2,0,1);
      glEnable(GL_MAP2_VERTEX_3);
      glMap2f(GL_MAP2_VERTEX_3,
         0,1,3*b->x.q3c.sizex,b->x.q3c.sizey,
         0,1,3,b->x.q3c.sizex,
         (float *)b->verts);
      glEvalMesh2(GL_FILL,
         0,cv_gl_curve_subdivide->i*(b->x.q3c.sizey-1)/2,
         0,cv_gl_curve_subdivide->i*(b->x.q3c.sizex-1)/2);
      glDisable(GL_MAP2_VERTEX_3);
      return;
   }
   
   for (f=b->faces,i=b->num_faces;i;i--,f++)
   {
      glBegin(GL_POLYGON);
      for (j=0;j<f->num_verts;j++)
      {
         v=&b->verts[f->verts[j]];
         glVertex3f(v->x,v->y,v->z);
      }
      glEnd();
   }
}

static void Flat_Ent(entity_t *e)
{
   glColor4f(e->eclass->color[0],e->eclass->color[1],e->eclass->color[2],1);
   if (e->num_brushes)
   {
      Wire_Ent(e);
   }
   else
   {
      glBegin(GL_QUAD_STRIP);
      glVertex3f(e->min.x,e->min.y,e->min.z);
      glVertex3f(e->min.x,e->min.y,e->max.z);
      glVertex3f(e->max.x,e->min.y,e->min.z);
      glVertex3f(e->max.x,e->min.y,e->max.z);
      glVertex3f(e->max.x,e->max.y,e->min.z);
      glVertex3f(e->max.x,e->max.y,e->max.z);
      glVertex3f(e->min.x,e->max.y,e->min.z);
      glVertex3f(e->min.x,e->max.y,e->max.z);
      glVertex3f(e->min.x,e->min.y,e->min.z);
      glVertex3f(e->min.x,e->min.y,e->max.z);
      glEnd();

      glBegin(GL_QUADS);
      glVertex3f(e->min.x,e->min.y,e->min.z);
      glVertex3f(e->max.x,e->min.y,e->min.z);
      glVertex3f(e->max.x,e->max.y,e->min.z);
      glVertex3f(e->min.x,e->max.y,e->min.z);
      
      glVertex3f(e->min.x,e->min.y,e->max.z);
      glVertex3f(e->min.x,e->max.y,e->max.z);
      glVertex3f(e->max.x,e->max.y,e->max.z);
      glVertex3f(e->max.x,e->min.y,e->max.z);
      glEnd();
   }
}

static void Tex_Draw(void)
{
   brush_t *b;
   bsel_t *bs;
   
   entity_t *e;
   esel_t *es;

   vec3_t *v;
   vsel_t *vs;

   face_t *f;
   fsel_t *fs;

   sel_t *s;
   
   int i,j;
   

   if (cv_render_flags->i&(RENDER_ENTITY|RENDER_BRUSH))
   {
      glEnable(GL_CULL_FACE);
      glEnable(GL_DEPTH_TEST);

      if (cv_render_flags->i&RENDER_BRUSH)
      {
         for (b=M.VBrushHead;b;b=b->VNext)
            if (b->drawn==1)
               Tex_Brush(b);
         FOR_EACH_SEL
         {
            if (!(s->flags&SEL_DRAW)) continue;
            for (bs=s->bsel;bs;bs=bs->Next)
               Tex_Brush(bs->Brush);
         }
         usetex(NULL);
      }

      if (cv_render_flags->i&RENDER_ENTITY)
      {
         for (e=M.VEntityHead;e;e=e->VNext)
            if (e->drawn==1 && e->uid!=M.WorldSpawn)
               Flat_Ent(e);
         FOR_EACH_SEL
         {
            if (!(s->flags&SEL_DRAW)) continue;
            for (es=s->esel;es;es=es->Next)
               Flat_Ent(es->Entity);
         }
      }

      glDisable(GL_CULL_FACE);
      if (cv_render_flags->i&RENDER_ENTITY_LINK)
         DrawLinks();
      glEnable(GL_CULL_FACE);
      
      glDepthFunc(GL_LEQUAL);
      glEnable(GL_BLEND);

      glColor4fv(C_BSEL_TEX);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
            Flat_BSel(bs->Brush);
      }
      
      glColor4fv(C_FSEL_TEX);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (fs=s->fsel;fs;fs=fs->Next)
         {
            b=fs->Brush;
            f=&b->faces[fs->num];
            glBegin(GL_POLYGON);
            for (j=0;j<f->num_verts;j++)
               glVertex3fv((float *)&(b->verts[f->verts[j]]));
            glEnd();
         }
      }
      
      glDisable(GL_BLEND);
      glDepthFunc(GL_LESS);
      
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_CULL_FACE);
   }

   if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
   {
      glColor4fv(C_BSEL_TEX_2);
      glBegin(GL_LINES);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
            Wire_Brush(bs->Brush);
      }
      glEnd();

      if (cv_render_flags->i&RENDER_BRUSH_VERT)
      {
         glColor4fv(C_VERT_TEX);
         P_glBegin(GL_POINTS);
         FOR_EACH_SEL
         {
            if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
               continue;
            for (bs=s->bsel;bs;bs=bs->Next)
            {
               b=bs->Brush;
               for (i=0,v=b->verts;i<b->num_verts;i++,v++)
               {
                  P_glVertex3fv((float *)v);
               }
            }
         }
         P_glEnd();
      }
   }

   if (cv_render_flags->i&RENDER_FACE_FOCAL)
   {
      glColor4fv(C_FSEL_TEX_2);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (fs=s->fsel;fs;fs=fs->Next)
         {
            b=fs->Brush;
            f=&b->faces[fs->num];
            glBegin(GL_LINE_LOOP);
            for (j=0;j<f->num_verts;j++)
               glVertex3fv((float *)&(b->verts[f->verts[j]]));
            glEnd();
         }
      }
      

      if (cv_render_flags->i&RENDER_FACE_VERT)
      {
         glColor4fv(C_VERT);
         P_glBegin(GL_POINTS);
         FOR_EACH_SEL
         {
            if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
               continue;
            for (fs=s->fsel;fs;fs=fs->Next)
            {
               b=fs->Brush;
               f=&b->faces[fs->num];
               for (j=0;j<f->num_verts;j++)
               {
                  P_glVertex3fv((float *)&b->verts[f->verts[j]]);
               }
            }
         }
         P_glEnd();
      }
   }
   
   glColor4fv(C_VSEL);
   P_glBegin(GL_POINTS);
   FOR_EACH_SEL
   {
      if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
         continue;
      for (vs=s->vsel;vs;vs=vs->Next)
      {
         P_glVertex3fv((float *)&(vs->Brush->verts[vs->num]));
      }
   }
   P_glEnd();


   if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
   {
      glColor4fv(C_ESEL);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (es=s->esel;es;es=es->Next)
            Wire_Ent(es->Entity);
      }
   }
}


static void Flat_Draw(void)
{
   brush_t *b;
   bsel_t *bs;
   
   entity_t *e;
   esel_t *es;

   vec3_t *v;
   vsel_t *vs;

   face_t *f;
   fsel_t *fs;

   sel_t *s;
   
   int i,j;
   

   if (cv_render_flags->i&(RENDER_ENTITY|RENDER_BRUSH))
   {
      glEnable(GL_CULL_FACE);
      glEnable(GL_DEPTH_TEST);

      if (cv_render_flags->i&RENDER_BRUSH)
      {
         for (b=M.VBrushHead;b;b=b->VNext)
            if (b->drawn==1)
               Flat_Brush(b);
         FOR_EACH_SEL
         {
            if (!(s->flags&SEL_DRAW)) continue;
            for (bs=s->bsel;bs;bs=bs->Next)
               Flat_Brush(bs->Brush);
         }
      }

      if (cv_render_flags->i&RENDER_ENTITY)
      {
         for (e=M.VEntityHead;e;e=e->VNext)
            if (e->drawn==1 && e->uid!=M.WorldSpawn)
               Flat_Ent(e);
         FOR_EACH_SEL
         {
            if (!(s->flags&SEL_DRAW)) continue;
            for (es=s->esel;es;es=es->Next)
               Flat_Ent(es->Entity);
         }
      }

      glDisable(GL_CULL_FACE);
      if (cv_render_flags->i&RENDER_ENTITY_LINK)
         DrawLinks();
      glEnable(GL_CULL_FACE);
      
      glDepthFunc(GL_LEQUAL);
      glEnable(GL_BLEND);

      glColor4fv(C_BSEL_TEX);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
            Flat_BSel(bs->Brush);
      }
      
      glColor4fv(C_FSEL_TEX);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (fs=s->fsel;fs;fs=fs->Next)
         {
            b=fs->Brush;
            f=&b->faces[fs->num];
            glBegin(GL_POLYGON);
            for (j=0;j<f->num_verts;j++)
               glVertex3fv((float *)&(b->verts[f->verts[j]]));
            glEnd();
         }
      }
      
      glDisable(GL_BLEND);
      glDepthFunc(GL_LESS);
      
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_CULL_FACE);
   }

   if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
   {
      glColor4fv(C_BSEL_TEX_2);
      glBegin(GL_LINES);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (bs=s->bsel;bs;bs=bs->Next)
            Wire_Brush(bs->Brush);
      }
      glEnd();

      if (cv_render_flags->i&RENDER_BRUSH_VERT)
      {
         glColor4fv(C_VERT_TEX);
         P_glBegin(GL_POINTS);
         FOR_EACH_SEL
         {
            if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
               continue;
            for (bs=s->bsel;bs;bs=bs->Next)
            {
               b=bs->Brush;
               for (i=0,v=b->verts;i<b->num_verts;i++,v++)
               {
                  P_glVertex3fv((float *)v);
               }
            }
         }
         P_glEnd();
      }
   }

   if (cv_render_flags->i&RENDER_FACE_FOCAL)
   {
      glColor4fv(C_FSEL_TEX_2);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (fs=s->fsel;fs;fs=fs->Next)
         {
            b=fs->Brush;
            f=&b->faces[fs->num];
            glBegin(GL_LINE_LOOP);
            for (j=0;j<f->num_verts;j++)
               glVertex3fv((float *)&(b->verts[f->verts[j]]));
            glEnd();
         }
      }
      

      if (cv_render_flags->i&RENDER_FACE_VERT)
      {
         glColor4fv(C_VERT);
         P_glBegin(GL_POINTS);
         FOR_EACH_SEL
         {
            if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
               continue;
            for (fs=s->fsel;fs;fs=fs->Next)
            {
               b=fs->Brush;
               f=&b->faces[fs->num];
               for (j=0;j<f->num_verts;j++)
               {
                  P_glVertex3fv((float *)&b->verts[f->verts[j]]);
               }
            }
         }
         P_glEnd();
      }
   }
   
   glColor4fv(C_VSEL);
   P_glBegin(GL_POINTS);
   FOR_EACH_SEL
   {
      if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
         continue;
      for (vs=s->vsel;vs;vs=vs->Next)
      {
         P_glVertex3fv((float *)&(vs->Brush->verts[vs->num]));
      }
   }
   P_glEnd();


   if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
   {
      glColor4fv(C_ESEL);
      FOR_EACH_SEL
      {
         if ((s->flags&(SEL_DRAW|SEL_SELECTED))!=(SEL_DRAW|SEL_SELECTED))
            continue;
         for (es=s->esel;es;es=es->Next)
            Wire_Ent(es->Entity);
      }
   }
}


static void CalcMatrix(gl_viewport_t *v)
{
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   if (v->mode&M_2D)
   {
      glOrtho(
         -v->sx/v->vp.zoom_amt,v->sx/v->vp.zoom_amt,
         -v->sy/v->vp.zoom_amt,v->sy/v->vp.zoom_amt,
         -4096,4096
         );
   }
   else
   {
      double ang=90;
      double y;

      ang=tan(ang/360.0*PI);
      y=ang*(v->sy/(double)v->sx);
      glFrustum(-ang,ang,-y,y,Z_NEAR,Z_FAR);
   }

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   {
      float rx,ry,rz;
      GetRotValues(&v->vp,&rx,&ry,&rz);
      glRotatef(-ry    ,0,0,1);
      glRotatef( rx-180,1,0,0);
      glRotatef( rz    ,0,0,1);
      glTranslatef(
         -v->vp.camera_pos.x,
         -v->vp.camera_pos.y,
         -v->vp.camera_pos.z);
   }

   {
      matrix_t m1,m2;
      glGetFloatv(GL_PROJECTION_MATRIX,(float *)m1);
      glGetFloatv(GL_MODELVIEW_MATRIX,(float *)m2);
      MultMatrix(m2,m1,v->m);
   }
}


static void DrawGrid(gl_viewport_t *v,double d)
{
   double y,y1,y2;
   double x,x1,x2;

   matrix_t m;

   if (d/2.0*v->vp.zoom_amt<10)
      return;

   glGetFloatv(GL_MODELVIEW_MATRIX,(float *)m);

   glPushMatrix();
   glLoadIdentity();
   
   y1=-v->sy/v->vp.zoom_amt;
   y1=floor((y1-d)/d)*d;
   
   y2=v->sy/v->vp.zoom_amt;
   y2=ceil((y2+d)/d)*d;
   
   x1=-v->sx/v->vp.zoom_amt;
   x1=floor((x1-d)/d)*d;
   
   x2=v->sx/v->vp.zoom_amt;
   x2=ceil((x2+d)/d)*d;

   m[3][0]-=floor(m[3][0]/d)*d;
   m[3][1]-=floor(m[3][1]/d)*d;

   x1+=m[3][0];
   x2+=m[3][0];
   y1+=m[3][1];
   y2+=m[3][1];

   glBegin(GL_LINES);
   for (y=y1;y<y2;y+=d)
   {
      glVertex2f(x1,y);
      glVertex2f(x2,y);
   }
   for (x=x1;x<x2;x+=d)
   {
      glVertex2f(x,y1);
      glVertex2f(x,y2);
   }
   glEnd();

   glPopMatrix();
}


/*
Keep a display list for each mode. If it's valid, we can call it to
render everything, and if it isn't, we'll recreate it. The *_valid flags
are cleared by the R_*_Update functions and whenever we need to render
stuff again.
*/
static int list[M_MAX_MODE];
static unsigned char valid[M_MAX_MODE];
static int last_render_flags;


static void (*(mfunc[M_MAX_MODE]))(void)=
{NULL,Wire_Draw,Tex_Draw,Flat_Draw};

void S_GL_Render(gl_viewport_t *v)
{
   sel_t *s;
   int i;
   int m;

   CalcMatrix(v);
   
   glClearColor(C_BACK[0],C_BACK[1],C_BACK[2],C_BACK[3]);
   if (v->mode==M_2D_WIRE || v->mode==M_3D_WIRE)
      glClear(GL_COLOR_BUFFER_BIT);
   else
      glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   if (!M.valid) return;

   if (last_render_flags!=cv_render_flags->i)
   {
      memset(valid,0,sizeof(valid));
      last_render_flags=cv_render_flags->i;
   }

   {
      if (v->mode&M_2D && v->vp.axis_aligned && v->grid_type==GRID_NORMAL)
      {
         glColor4fv(C_GRID_SMALL);
         DrawGrid(v,cv_edit_snap_size->f);
         glColor4fv(C_GRID_LARGE);
         DrawGrid(v,cv_edit_snap_size->f*4);
      }
      
      if (cv_render_flags->i&RENDER_BRUSH)
      {
         brush_t *b;
         bsel_t *bs;
      
         for (b=M.VBrushHead;b;b=b->VNext)
            b->drawn=1;
         if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
         {
            FOR_EACH_SEL
            {
               for (bs=s->bsel;bs;bs=bs->Next)
                  bs->Brush->drawn=2;
            }
         }
      }

      if (cv_render_flags->i&RENDER_ENTITY)
      {
         entity_t *e;
         esel_t *es;
      
         for (e=M.VEntityHead;e;e=e->VNext)
            e->drawn=1;
         if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
         {
            FOR_EACH_SEL
            {
               for (es=s->esel;es;es=es->Next)
                  es->Entity->drawn=2;
            }
         }
      }

      m=v->mode&M_MODE;
      if (cv_gl_use_display_lists->i)
      {
         if (valid[m])
            glCallList(list[m]);
         else
         {
            glNewList(list[m],GL_COMPILE_AND_EXECUTE);
            mfunc[m]();
            glEndList();
            valid[m]=1;
         }
      }
      else
         mfunc[m]();
   }
}


static void SetMode(gl_viewport_t *v,int mode)
{
   v->mode=mode;
   if (mode&M_2D)
      v->vp.axis_aligned=1;
   else
      v->vp.axis_aligned=0;
}


void S_GL_Resize(gl_viewport_t *v)
{
   glViewport(0,0,v->sx,v->sy);
}

void S_GL_Init(gl_viewport_t *v)
{
   SetMode(v,M_3D_WIRE);
   v->grid_type=GRID_NORMAL;

   glPointSize(2);
   glFrontFace(GL_CW);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
   
   S_GL_Resize(v);
}


void S_GL_SelectionWindow_DrawBox(int x0,int y0,int x1,int y1)
{
   gl_viewport_t *v=S_GL_vport;
   
   S_GL_Render(v);
   
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0,v->sx,v->sy,0, -1,1);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glColor4fv(C_SELECTION_BOX);
   glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
   glRectf(x0,y0,x1,y1);
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}


int S_R_Trans_Vert(vec3_t v,vec3_t *sv)
{
   gl_viewport_t *vp=S_GL_vport;
   double w;
   
   sv->x=vp->m[0][0]*v.x+
         vp->m[1][0]*v.y+
         vp->m[2][0]*v.z+
         vp->m[3][0];
   sv->y=vp->m[0][1]*v.x+
         vp->m[1][1]*v.y+
         vp->m[2][1]*v.z+
         vp->m[3][1];
   sv->z=vp->m[0][2]*v.x+
         vp->m[1][2]*v.y+
         vp->m[2][2]*v.z+
         vp->m[3][2];
   w=vp->m[0][3]*v.x+
     vp->m[1][3]*v.y+
     vp->m[2][3]*v.z+
     vp->m[3][3];

   if (sv->z<0 && !(vp->mode&M_2D)) return 0;
     
   if (w)
   {
      sv->x/=w;
      sv->y/=w;
   }

   return 1;
}

int S_R_Trans_SVert(vec3_t sv,vec3_t *v)
{
   matrix_t i;
   vec3_t p;
   
   if (!(S_GL_vport->mode&M_2D))
      return 0;

   InvMatrix(S_GL_vport->m,i);
   
   p.x=i[0][0]*sv.x+
       i[1][0]*sv.y+
       i[3][0];
   p.y=i[0][1]*sv.x+
       i[1][1]*sv.y+
       i[3][1];
   p.z=i[0][2]*sv.x+
       i[1][2]*sv.y+
       i[3][2];

   *v=p;
   
   return 1;
}

int S_R_GetVector(vec3_t spos,vec3_t *org,vec3_t *dir)
{
   vec3_t p1,p2,t;
   double p1w,p2w;
   gl_viewport_t *v=S_GL_vport;
   matrix_t i;

   if (v->mode&M_2D)
   {
      p1.x=spos.x;
      p1.y=spos.y;
      p1.z=0;
      p1w=1;
      p2.x=spos.x;
      p2.y=spos.y;
      p2.z=1;
      p2w=1;
   }
   else
   {
      p1.x=p1.y=0;
      p1.z=-2*Z_FAR*Z_NEAR/(Z_FAR-Z_NEAR);
      p1w=0;
      
      p2.x=spos.x;
      p2.y=spos.y;
      p2.z=(Z_FAR+Z_NEAR-2*Z_FAR*Z_NEAR)/(Z_FAR-Z_NEAR);
      p2w=1;
   }
   
   InvMatrix(v->m,i);

   t=p1;
   p1.x=i[0][0]*t.x+
        i[1][0]*t.y+
        i[2][0]*t.z+
        i[3][0]*p1w;
   p1.y=i[0][1]*t.x+
        i[1][1]*t.y+
        i[2][1]*t.z+
        i[3][1]*p1w;
   p1.z=i[0][2]*t.x+
        i[1][2]*t.y+
        i[2][2]*t.z+
        i[3][2]*p1w;
        
   t=p2;
   p2.x=i[0][0]*t.x+
        i[1][0]*t.y+
        i[2][0]*t.z+
        i[3][0]*p2w;
   p2.y=i[0][1]*t.x+
        i[1][1]*t.y+
        i[2][1]*t.z+
        i[3][1]*p2w;
   p2.z=i[0][2]*t.x+
        i[1][2]*t.y+
        i[2][2]*t.z+
        i[3][2]*p2w;
      
   p2.x-=p1.x;
   p2.y-=p1.y;
   p2.z-=p1.z;
   Normalize(&p2);

   /*
   If it's a 2d viewport, move back a bit so a trace doesn't start inside
   the map (i.e. try to ignore position along trace vector, only confusing).
   */
   if (v->mode&M_2D)
   {
      p1.x-=p2.x*4096;
      p1.y-=p2.y*4096;
      p1.z-=p2.z*4096;
   }
   
   *org=p1;
   *dir=p2;
      
   return 1;
}


int S_R_IsNear(vec3_t sv1,vec3_t sv2)
{
   if (fabs((sv1.x-sv2.x)*S_GL_vport->sx/2)<8 &&
       fabs((sv1.y-sv2.y)*S_GL_vport->sy/2)<8)
      return 1;

   return 0;
}

int S_R_IsInBox(vec3_t sv,vec3_t min,vec3_t max)
{
   if (sv.x>min.x && sv.y>min.y &&
       sv.x<max.x && sv.y<max.y)
      return 1;
   return 0;
}


int S_R_PreferSolid(void)
{
   switch (S_GL_vport->mode&M_MODE)
   {
   default:
   case M_WIRE:
      return 0;

   case M_FLAT:
   case M_TEXTURE:
      return 1;
   }
}


void S_R_EntityUpdate(entity_t *e)
{
   (void)e;
   memset(valid,0,sizeof(valid));
}
void S_R_EntityFree(entity_t *e)
{
   (void)e;
   memset(valid,0,sizeof(valid));
}

void S_R_FaceUpdate(struct face_s *f)
{
static vec3_t lvec={0.758,0.530,0.379};

   if (!f->r)
   {
      f->r=Q_malloc(sizeof(r_face_t));
      if (!f->r)
         S_Abort("S_R_FaceUpdate","Out of memory!");
      memset(f->r,0,sizeof(r_face_t));
   }
   f->r->t=S_GL_LoadTexture(f->tex.name);
   f->r->light=(DotProd(lvec,f->normal)+2)/3.0;
   
   memset(valid,0,sizeof(valid));
}
void S_R_FaceFree(struct face_s *f)
{
   if (f->r)
   {
      Q_free(f->r);
      f->r=NULL;
   }
   
   memset(valid,0,sizeof(valid));
}

void S_R_BrushUpdate(struct brush_s *b)
{
   (void)b;
   memset(valid,0,sizeof(valid));
}
void S_R_BrushFree(struct brush_s *b)
{
   (void)b;
   memset(valid,0,sizeof(valid));
}


void S_R_SelUpdate(void)
{
   memset(valid,0,sizeof(valid));
}


void S_R_Map_Free(void)
{
   memset(valid,0,sizeof(valid));
}


/*

(x,y,z) => (x,y)==(a,b)

v=(a,b,0)+s(0,0,1)

(x,y,z) => (x/z,y/z)==(a,b)

v=(x,y,1)+s(f,g,h)

(x+sf)/(1+sh)=x
(y+sg)/(1+sh)=y

x+sf=x+xhs
y+sg=y+yhs

sf=xhs
sg=yhs

v=(x,y,1)+s(xh,yh,h)
v=(x,y,1)+sh(x,y,1)=(sh+1)(x,y,1)

*/


/*

A

v=(a b c)

v*A=t

t=(x y z w)

x/w=s
y/w=t

t=(s t z 1)

(a b c)*A=(s t z 1)

1=-c
z=-c*(far+nr)/(far-nr)-(2*far*nr)/(far-nr)

z=((far+nr)-(2*far*nr))/(far-nr)


t=(2s 2t p 2)
(a b c)*A=(2s 2t p 2)

c=-2
2s=fa*a
2t=f*b
p=2*(far+nr)/(far-nr)-(2*far*nr)/(far-nr)
=(2*(far+nr)-2*far*nr)/(far-nr)

(s/fa t/f z)
(2s/fa 2t/f p)

(nr*(far+nr)/(far-nr)-2*far*nr/(far-nr))/nr=
(far+nr-2*far)/(far-nr)=-1

(far*(far+nr)/(far-nr)-2*far*nr/(far-nr))/far=
(far+nr-2*nr)/(far-nr)=1

*/


#include "cmds.h"

static void C_toggle_texture(void)
{
   if ((S_GL_vport->mode&M_MODE)==M_TEXTURE)
      SetMode(S_GL_vport,(S_GL_vport->mode&~M_MODE)|M_WIRE);
   else
      SetMode(S_GL_vport,(S_GL_vport->mode&~M_MODE)|M_TEXTURE);
   Cmd_SetRedraw(1);
}
static void C_toggle_flat(void)
{
   if ((S_GL_vport->mode&M_MODE)==M_FLAT)
      SetMode(S_GL_vport,(S_GL_vport->mode&~M_MODE)|M_WIRE);
   else
      SetMode(S_GL_vport,(S_GL_vport->mode&~M_MODE)|M_FLAT);
   Cmd_SetRedraw(1);
}

static void C_toggle_3d(void)
{
   if (S_GL_vport->mode&M_2D)
      SetMode(S_GL_vport,(S_GL_vport->mode&M_MODE)|M_3D);
   else
      SetMode(S_GL_vport,(S_GL_vport->mode&M_MODE)|M_2D);
   Cmd_SetRedraw(1);
}

static void C_toggle_grid(void)
{
   switch (S_GL_vport->grid_type)
   {
   default:
   case GRID_NONE:
      S_GL_vport->grid_type=GRID_NORMAL;
      break;
   case GRID_NORMAL:
      S_GL_vport->grid_type=GRID_NONE; /* TODO: make GRID_SPECIAL work */
      break;
   case GRID_SPECIAL:
      S_GL_vport->grid_type=GRID_NONE;
      break;
   }
   Cmd_SetRedraw(1);
}

#include "s_3d.h"
#include "s_sys.h"
static void C_profile(void)
{
   gl_viewport_t *v=S_GL_vport;
   int i;
   double et;
#define NUM 100

   v->vp.camera_pos.x=75;
   v->vp.camera_pos.y=500;
   v->vp.camera_pos.z=150;

   v->vp.rot.z=0;

   et=Sys_Time();
   for (i=0;i<NUM;i++)
   {
      v->vp.rot.z+=11;
      S_UpdateCurViewport();
   }
   et=Sys_Time()-et;
   printf("gl.profile: %g secs, %g secs/frame\n",
      et,et/(double)NUM);
#undef NUM
}

cmd_reg_t gl_cmds[]={
{"gl.toggle_3d"     ,C_toggle_3d     ,{"View/Toggle 3d"     ,"f"},{},"Toggle 2d/3d view in current viewport."},
{"gl.toggle_texture",C_toggle_texture,{"View/Toggle texture","t"},{},"Toggle textured view in current viewport."},
{"gl.toggle_flat"   ,C_toggle_flat   ,{"View/Toggle flat"   ,"alt+h"},{},"Toggle flat-shaded view in current viewport."},
{"gl.toggle_grid"   ,C_toggle_grid   ,{"View/Toggle grid"   ,"g"},{},"Toggle grid type in current viewport."},

{"gl.profile",C_profile,{"Misc/Debug/Profile","control+alt+p"},{},"Profile the OpenGL renderer"},
};

void S_GL_InitRender(void)
{
   int i;
   
   Cmd_Register(NULL,gl_cmds,sizeof(gl_cmds));
   cv_gl_curve_subdivide=
      Cfg_DefVar("gl.curve_subdivide","3",
         "How many times should curves be subdivided (higher values give "
         "better quality, but lower speed).");
   cv_gl_use_display_lists=
      Cfg_DefVar("gl.use_display_lists","1",
         "Use display lists to speed up rendering.");

   for (i=1;i<M_MAX_MODE;i++)
      list[i]=glGenLists(1);
   memset(valid,0,sizeof(valid));
   
   {
      char buf[128],cbuf[64];
      int i;
      config_var_t *cv;
      float f[4];

      for (i=0;i<C_MAX;i++)
      {
         sprintf(buf,"gl.color.%s",colors[i].name);
         sprintf(cbuf,"%0.3f %0.3f %0.3f %0.3f",
            colors[i].c[0],colors[i].c[1],colors[i].c[2],colors[i].c[3]);
         cv=Cfg_DefVar(buf,cbuf,colors[i].desc);
         if (sscanf(cv->s,"%f %f %f %f",&f[0],&f[1],&f[2],&f[3])==4)
         {
            colors[i].c[0]=f[0];
            colors[i].c[1]=f[1];
            colors[i].c[2]=f[2];
            colors[i].c[3]=f[3];
         }
      }
   }
}

void S_GL_DoneRender(void)
{
   int i;
   for (i=1;i<M_MAX_MODE;i++)
      glDeleteLists(list[i],1);
}

