/*
types.h file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#ifndef TYPES_H
#define TYPES_H

/* Base types */

typedef float scalar_t;

typedef struct vec3_s
{
   scalar_t x;
   scalar_t y;
   scalar_t z;
} vec3_t;

typedef struct vec2_s
{
   scalar_t x;
   scalar_t y;
} vec2_t;

typedef scalar_t matrix_t[4][4];

typedef struct edge_s
{
   int startvertex;            /* ID of the start vertex */
   int endvertex;              /* ID of the end vertex */
} edge_t;

typedef struct
{
   vec3_t n;
   scalar_t dist;
} plane_t;


/* Map objects (brush related stuff is in brush.h)

The map objects are brushes, entities, and groups. These are the objects the
make the map, and they're the ones that the undo system tracks.

Map objects are stored in linked lists with heads in the M structure. Each
objects has a Next and a Last field, a uid (a unique number used to identify
it in the undo system (and elsewhere)), and a real field, which should be set
when the object is linked into the main lists and clear otherwise (handled
automatically).

General function for managing them are in brush.c, entity.c and group.c . The
functions should be (where x is B, E or G):

x_t *x_New();
  Allocates a new, empty object and returns a pointer to it (or NULL on
  failure). Doesn't link it or set the uid.

void x_Free(x_t *X)
  Frees X. Doesn't unlink it.

void x_FreeC(x_t *X)
  Frees the contents of X, but not X itself (so you can allocate X e.g.
  statically). Doesn't unlink it.

void x_Link(x_t *X)
  Gives X a uid and links it into the main map data lists. Does a U_xxxAdd().
  Sets X->real. Does any work necessary to get the object ready for the main
  lists (i.e. allocate renderer data).

void x_Unlink
  Unlinks X from the main map data lists. Does a U_xxxDelete(). Doesn't free
  it or clear the uid. Clears X->real.

x_t *x_Duplicate(x_t *dest,x_t *src)
  Creates a duplicate of src in dest. If dest is NULL, allocate a new one.
  Returns the new object or NULL.

int x_Size(x_t *X)
  Returns the size of all stuff allocated by X. The size of the whole object
  X is x_Size(x)+sizeof(x_t). Warning: not absolutely accurate right now,
  only used to figure out when to free undo information.

void x_Update(x_t *X)
  This updates all fields in X to make it ready for rendering. Specifically,
  it makes sure the renderer is up-to-date on the object, and for brushes,
  it calculates normals and center, and for entities, it calculates center,
  min, max and nbrushes. Undo_Update() calls it for all modified objects, so
  interactive commands should call Undo_Update() if it needs to render after
  modifying stuff. If you add/delete anything, or change something that might
  affect what's visible, you should call Map_UpdateVLists(), which recreates
  the M.V{Brush,Entity}Head lists. Map_Update() updates everything (all
  objects and the visible lists).


Note that you should _NOT_ allocate, free, link or unlink any map objects
manually, it should all be done in these functions (there might be renderer
or game hooks, or new brush types with extra data to be duplicated, or other
things. Also, there are a few exceptions in internal code (undo system, map
clearing)).

Each map object has a misc field. It can be used as a general flag in
commands to keep track of stuff (e.g. the edit.move_ commands uses it to keep
track of which brushes to move by their vertices, and which will be moved by
a selected entity). The field should always be 0 for all map objects when
you're done.

*/

/*
This should be used to reference map objects (entities, brushes, groups).
Editing or undoing stuff might move them around, and for the clipboard system
stuff might not be around when you paste things.

The only things that maintain pointers to map objects are the linked lists of
map data, and the selection system. Thus, the linked list pointers (Next,
Last) must always be kept up to date with changes, and the selection must be
cleared if things move around in memory. The visible lists should be rebuilt
if necessary.

Other systems (e.g. the clipboard stuff) might also use pointers for linked
lists internally, but that's their problem.

0 indicates no object.
*/
typedef unsigned int map_uid_t;


/*
   Group flags (8 bits, therefore 8 flags)
*/
#define GROUP_WORLD         0x01 /* The WorldSpawn Group */
#define GROUP_HIDDEN        0x02 /* DON'T display */
#define GROUP_SMARTSELECT   0x04 /* Smart Selecting (selecting 1 of the group
                                    selects all) */
typedef struct group_s
{
   struct group_s *Next;
   struct group_s *Last;
   map_uid_t uid;
   int misc;
   unsigned char real;

   unsigned char color;
   unsigned char flags;
   char groupname[64];

   map_uid_t parent,child,cnext,clast;

   unsigned char hidden; /* is it really hidden */
   char fullname[256]; /* full name, i.e. 'parent.parent.etc.name' */
   /* TODO: are dots safe? people might have used them in names before */
} group_t;


typedef struct entity_s
{
   struct entity_s *Last;
   struct entity_s *Next;
   struct entity_s *VLast;
   struct entity_s *VNext;
   map_uid_t uid;
   unsigned char real;

   int misc;
   
   char   **key;
   char   **value;
   int      num_keys;

   map_uid_t Group;
   
   vec3_t center;
   vec3_t min,max;
   int num_brushes;

   struct entity_s *disp;

   struct eclass_s *eclass;

   int drawn; /* often the only renderer field, avoid having to allocate
                 another structure */
   struct S_r_entity_s *r;
} entity_t;


/* Structures used for selecting stuff */

typedef struct esel_s
{
   struct esel_s *Next;
   struct esel_s *Last;
   
   entity_t *Entity;
} esel_t;

typedef struct bsel_s
{
   struct bsel_s *Next;
   struct bsel_s *Last;

   struct brush_s *Brush;
} bsel_t;

typedef struct vsel_s
{
   struct vsel_s *Next;
   struct vsel_s *Last;

   struct brush_s *Brush;
   int num;
} vsel_t;

typedef struct fsel_s
{
   struct fsel_s *Next;
   struct fsel_s *Last;
   
   struct brush_s *Brush;
   int num;
} fsel_t;


/* Texture definition */

typedef struct texdef_s
{
   char name[MAX_TEXTURE_NAME];
   struct G_texdef_s *g;
} texdef_t;


/* Viewport structure */
typedef struct viewport_s
{
   /* When this is 1, the other values are valid and the viewport can be
   rendered. Viewports start off with this set to 0 so the renderer won't try
   to draw stuff with e.g. zoom_amt==0 */
   int     active;

	vec3_t  camera_pos;
   float   zoom_amt;
   int     axis_aligned;
   vec3_t  rot;
} viewport_t;


/* Selection structure, basically a bunch of objects */
typedef struct
{
   int flags;
#define SEL_DRAW       1   /* should it be drawn at all? */
#define SEL_SELECTED   2   /* should the objects be drawn 'selected'? */

	/* Vertex selection */
	int     num_vselected;
	vsel_t *vsel;
	/* Face selection */
	int     num_fselected;
	fsel_t *fsel;
	/* Brush selection */
	int     num_bselected;
	bsel_t *bsel;
	/* Entity selection */
	int     num_eselected;
	esel_t *esel;
} sel_t;


#endif

