#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>

/* these are redefined in defines.h */
#undef TRUE
#undef FALSE

#include "defines.h"
#include "types.h"

#include "g_3d_all.h"
#include "s_render.h"
#include "s_trans.h"

#include "g_3d.h"

#include "brush.h"
#include "camera.h"
#include "config.h"
#include "entity.h"
#include "geom.h"
#include "s_error.h"
#include "status.h"
#include "uid.h"


typedef struct svec_s
{
   int x,y;
   int onscreen;
} svec_t;


typedef struct S_r_entity_s
{
   int      drawn;
} r_entity_t;

typedef struct S_r_brush_s
{
   int num_alloc_verts;
   vec3_t  *tverts;
   svec_t  *sverts;
   int      drawn;
   svec_t   scenter;
} r_brush_t;

typedef struct S_r_face_s
{
   svec_t  scenter;
} r_face_t;


/* TODO:
use Q_{malloc,free} in these?
handle errors?
*/
void S_R_EntityUpdate(entity_t *e)
{
   if (!e->r)
   {
      e->r=malloc(sizeof(r_entity_t));
      if (!e->r) S_Abort("S_R_EntityUpdate","Out of memory!");
      memset(e->r,0,sizeof(r_entity_t));
   }
}
void S_R_EntityFree(entity_t *e)
{
   if (e->r) free(e->r);
   e->r=NULL;
}

void S_R_BrushUpdate(struct brush_s *b)
{
   if (!b->r)
   {
      b->r=malloc(sizeof(r_brush_t));
      if (!b->r) S_Abort("S_R_BrushUpdate","Out of memory!");
      memset(b->r,0,sizeof(r_brush_t));
   }
   if (b->r->num_alloc_verts!=b->num_verts)
   { /* TODO: handle b->num_verts==0 */
      b->r->tverts=realloc(b->r->tverts,sizeof(vec3_t)*b->num_verts);
      b->r->sverts=realloc(b->r->sverts,sizeof(svec_t)*b->num_verts);
      if (!b->r->tverts || !b->r->sverts)
         S_Abort("S_R_BrushUpdate","Out of memory");
      b->r->num_alloc_verts=b->num_verts;
   }
}

void S_R_BrushFree(struct brush_s *b)
{
   if (b->r)
   {
      free(b->r->tverts);
      free(b->r->sverts);
      free(b->r);
      b->r=NULL;
   }
}

void S_R_FaceUpdate(struct face_s *f)
{
   if (!f->r)
   {
      f->r=malloc(sizeof(r_face_t));
      if (!f->r) S_Abort("S_R_FaceUpdate","Out of memory!");
      memset(f->r,0,sizeof(r_face_t));
   }
}

void S_R_FaceFree(struct face_s *f)
{
   if (f->r) free(f->r);
   f->r=NULL;
}



static config_var_t
   *cv_depth_clip,
   *cv_fov,
   *cv_grid_font;


/* The fov for 3d viewports */
#define ALPHA ((cv_fov->f/2)*PI/180.0)


static void GenerateMatrix(matrix_t m,g_viewport_t *v)
{
   int      i, j;
   matrix_t mmain,         /* Matrices used for rotations, tranformations */
            rottrans,      /* between spaces */
            trans,
            rot,
            canon;
   float    ar;            /* Aspect ration of screen (width/height) */
   float    rx,ry,rz;


   ar=(float)(v->sx) /
      (float)(v->sy);

   /* First build main transformation matrix */
   InitMatrix(mmain);

   InitMatrix(trans);
   trans[0][3] = -v->vp.camera_pos.x;
   trans[1][3] = -v->vp.camera_pos.y;
   trans[2][3] = -v->vp.camera_pos.z;

   GetRotValues(&v->vp,&rx,&ry,&rz);
   GenerateRotMatrix(rot,rx,ry,rz);

   InitMatrix(canon);
   if (v->mode != M_2D_WIRE)
   {
      canon[0][0] = (float)1  / (cv_depth_clip->f * tan(ALPHA));
      canon[1][1] = (float)ar / (cv_depth_clip->f * tan(ALPHA));
      canon[2][2] = (float)1  / cv_depth_clip->f;
   }

   /* Multiply Matricies into one tranformation matrix */
   MultMatrix(rot,   trans,    rottrans);
   MultMatrix(canon, rottrans, mmain);

   /* Copy into result matrix */
   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
         m[i][j] = mmain[i][j];
}

static void GenerateIMatrix(matrix_t m,g_viewport_t *v)
{
   int      i, j;
   matrix_t mmain,         /* Matrices used for rotations, tranformations */
            rottrans,      /* between spaces */
            trans,
            rot,
            canon;
   float    ar;            /* Aspect ration of screen (width/height) */
   float    rx,ry,rz;


   ar=(float)(v->sx) /
      (float)(v->sy);

   /* First build main transformation matrix */
   InitMatrix(mmain);

   InitMatrix(trans);
   trans[0][3] = v->vp.camera_pos.x;
   trans[1][3] = v->vp.camera_pos.y;
   trans[2][3] = v->vp.camera_pos.z;

   GetRotValues(&v->vp,&rx,&ry,&rz);
   GenerateIRotMatrix(rot,-rx,-ry,-rz);

   InitMatrix(canon);
   if (v->mode != M_2D_WIRE)
   {
      canon[0][0] = (float)(cv_depth_clip->f * tan(ALPHA));
      canon[1][1] = (float)(cv_depth_clip->f * tan(ALPHA)) / ar;
      canon[2][2] = (float)cv_depth_clip->f;
   }

   /* Multiply Matricies into one tranformation matrix */
   MultMatrix(trans,      rot,    rottrans);
   MultMatrix(rottrans, canon,       mmain);

   /* Copy into result matrix */
   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
         m[i][j] = mmain[i][j];
}


GdkColormap *S_colormap;
static GdkFont *grid_font;
static GdkGC *gc;


static GdkColor colors[]=
{
#define  C_BACK             (&colors[ 0])
{0,0x0000,0x0000,0x0000},

#define  C_GRID_1           (&colors[ 1])
{0,0x0000,0x0000,0x8000},
#define  C_GRID_2           (&colors[ 2])
{0,0x0000,0x0000,0xb000},
#define  C_GRID_TEXT        (&colors[ 3])
{0,0xffff,0xffff,0xffff},

#define  C_BRUSH_SELECT     (&colors[ 4])
{0,0xffff,0xffff,0x0000},
#define  C_BRUSH_FOCAL_SUB  (&colors[ 5])
{0,0xffff,0x0000,0xffff},
#define  C_BRUSH_FOCAL      (&colors[ 6])
{0,0xffff,0x0000,0x0000},

#define  C_FACE_SELECT      (&colors[ 7])
{0,0xffff,0xffff,0x0000},
#define  C_FACE_FOCAL       (&colors[ 8])
{0,0x0000,0xffff,0x0000},

#define  C_VERT             (&colors[ 9])
{0,0x0000,0xffff,0xffff},
#define  C_VERT_SEL         (&colors[10])
{0,0xffff,0x0000,0x0000},

#define  C_BORDER_ACTIVE    (&colors[11])
{0,0xffff,0xffff,0xffff},
#define  C_BORDER_OTHER     (&colors[12])
{0,0x8000,0x8000,0x8000},

#define  C_PTS_DIST         (&colors[13])
{0,0xffff,0xffff,0x0000},
#define  C_PTS_DIST_TEXT    (&colors[14])
{0,0xffff,0xffff,0xffff},

#define  C_PTS_BRUSH_END    (&colors[15])
{0,0xffff,0x0000,0x0000},
#define  C_PTS_BR           (&colors[16])
{0,0xffff,0xffff,0x0000},

#define  C_PTS_LEAK_START   (&colors[17])
{0,0x0000,0xffff,0x0000},
#define  C_PTS_LEAK_END     (&colors[18])
{0,0x0000,0x0000,0xffff},
#define  C_PTS_LEAK         (&colors[19])
{0,0xffff,0x0000,0x0000},

#define  C_AXIS_X           (&colors[20])
{0,0xffff,0x0000,0x0000},
#define  C_AXIS_Y           (&colors[21])
{0,0x0000,0xffff,0x0000},
#define  C_AXIS_Z           (&colors[22])
{0,0x0000,0xffff,0xffff},

#define  C_ENTITY_FOCAL     (&colors[23])
{0,0x8000,0x8000,0x8000},
#define  C_ENTITY_SELECT    (&colors[24])
{0,0xffff,0x8000,0x8000},


#define  C_SELECTION_BOX    (&colors[25])
{0,0x2000,0x2000,0xffff},

#define C_MAX 26
};


static GdkColor *GroupColor(int num)
{
   return C_GRID_TEXT;
}


/* Transform vec3_t s to vec3_t d using matrix_t m */
#define Transform(d,s,m) \
({ \
   d.x=m[0][0]*s.x+ \
       m[0][1]*s.y+ \
       m[0][2]*s.z+ \
       m[0][3]; \
   d.y=m[1][0]*s.x+ \
       m[1][1]*s.y+ \
       m[1][2]*s.z+ \
       m[1][3]; \
   d.z=m[2][0]*s.x+ \
       m[2][1]*s.y+ \
       m[2][2]*s.z+ \
       m[2][3]; \
})


static int dx,dy;
static g_viewport_t *cur_v;
static GdkColor *cur_color;


/* Wrappers for GDK calls */
static void SetColor(GdkColor *col)
{
   if (cur_color->pixel!=col->pixel)
   {
      gdk_gc_set_foreground(gc,col);
      cur_color=col;
   }
}

static void DrawLine(int x0,int y0,int x1,int y1)
{
   gdk_draw_line(cur_v->pic,gc,
      x0,y0,x1,y1);
}

static void DrawStr(int x,int y,GdkFont *f,const char *str)
{
   gdk_draw_string(cur_v->pic,f,gc,
      x,y,str);
}

static void DrawDot(svec_t s)
{
   if (!s.onscreen) return;
   gdk_draw_arc(cur_v->pic,gc,0,
      s.x-1,s.y-1,3,3,0,360*64);
}

static void DrawVertex(svec_t s)
{
   if (!s.onscreen) return;
   gdk_draw_rectangle(cur_v->pic,gc,1,
      s.x-1,s.y-1,3,3);
}

static void DrawBox(svec_t s,int size)
{
   if (!s.onscreen) return;
   gdk_draw_rectangle(cur_v->pic,gc,0,
      s.x-size/2,s.y-size/2,size,size);
}


/* Include clipping functions */
/* TODO: do some better way? keeping it in a separate file makes
a mess of the names. Too bad there isn't a 'module'-scope thing in
c. */
#include "g_3d_clip.c"


/* Grid drawing */
static void DrawGrid(int snap)
{
   int x1;
   vec3_t m;
   int ofs;
   int s;
   int ds;

   float pos,delta;

   int i;


   delta=(float)snap*cur_v->vp.zoom_amt;
   if (delta<=4)
      return;

/*   if (delta<56)
      c=0;*/

/* X grid */
   x1= (((float)-dx)/cur_v->vp.zoom_amt);

   Move(&cur_v->vp,MOVE_RIGHT,&m,1);
   if (m.x)
   {
      ofs=cur_v->vp.camera_pos.x;
      i=m.x;
   }
   else
   if (m.y)
   {
      ofs=cur_v->vp.camera_pos.y;
      i=m.y;
   }
   else
   {
      ofs=cur_v->vp.camera_pos.z;
      i=m.z;
   }
   x1+=i*ofs;
   x1-=x1%snap;

   pos=(x1-i*ofs)*cur_v->vp.zoom_amt + dx;

   x1*=i;

   s=x1;
   if (i==-1)
      ds=-snap;
   else
      ds= snap;

   while (pos<0)
   {
      s+=ds;
      pos+=delta;
      i++;
   }

   while (pos<cur_v->sx)
   {
      DrawLine(pos,0,pos,cur_v->sy-1);
      pos+=delta;
      s+=ds;
   }

/* Y grid */
   x1=-(((float)-dy)/cur_v->vp.zoom_amt);

   Move(&cur_v->vp,MOVE_DOWN,&m,1);
   if (m.x)
   {
      ofs=cur_v->vp.camera_pos.x;
      i=m.x;
   }
   else
   if (m.y)
   {
      ofs=cur_v->vp.camera_pos.y;
      i=m.y;
   }
   else
   {
      ofs=cur_v->vp.camera_pos.z;
      i=m.z;
   }
   x1-=i*ofs;
   x1-=x1%snap;

   pos=-(x1+i*ofs)*cur_v->vp.zoom_amt + dy;

   x1*=-i;

   s=x1;
   if (i==-1)
      ds=-snap;
   else
      ds= snap;

   while (pos<0)
   {
      s+=ds;
      pos+=delta;
   }

   while (pos<cur_v->sy)
   {
      DrawLine(0,pos,cur_v->sx-1,pos);
      pos+=delta;
      s+=ds;
   }
}

static void DrawGridNum(void)
{
   int x1;
   vec3_t m;
   int ofs;
   int s;
   int ds;

   float pos,delta;

   int i;

   char temp[16];
   int len;

   int snap;


   delta=64/cur_v->vp.zoom_amt;
   for (snap=1;snap<delta;snap<<=1) ;


   delta=(float)snap*cur_v->vp.zoom_amt;

/* X grid */
   x1= (((float)-dx)/cur_v->vp.zoom_amt);

   Move(&cur_v->vp,MOVE_RIGHT,&m,1);
   if (m.x)
   {
      ofs=cur_v->vp.camera_pos.x;
      i=m.x;
   }
   else
   if (m.y)
   {
      ofs=cur_v->vp.camera_pos.y;
      i=m.y;
   }
   else
   {
      ofs=cur_v->vp.camera_pos.z;
      i=m.z;
   }
   x1+=i*ofs;
   x1-=x1%snap;

   pos=(x1-i*ofs)*cur_v->vp.zoom_amt + dx;

   x1*=i;

   s=x1;
   if (i==-1)
      ds=-snap;
   else
      ds= snap;

   while (pos<0)
   {
      s+=ds;
      pos+=delta;
      i++;
   }

   while (pos<cur_v->sx)
   {
      sprintf(temp,"%i",s);
      len=gdk_string_width(grid_font,temp);
      i=pos-len/2;
      if ((i>0) && (i+len<cur_v->sx))
      {
         DrawStr(i,2+gdk_string_height(grid_font,temp),grid_font,temp);
      }

      pos+=delta;
      s+=ds;
   }

/* Y grid */
   x1=-(((float)-dy)/cur_v->vp.zoom_amt);

   Move(&cur_v->vp,MOVE_DOWN,&m,1);
   if (m.x)
   {
      ofs=cur_v->vp.camera_pos.x;
      i=m.x;
   }
   else
   if (m.y)
   {
      ofs=cur_v->vp.camera_pos.y;
      i=m.y;
   }
   else
   {
      ofs=cur_v->vp.camera_pos.z;
      i=m.z;
   }
   x1-=i*ofs;
   x1-=x1%snap;

   pos=-(x1+i*ofs)*cur_v->vp.zoom_amt + dy;

   x1*=-i;

   s=x1;
   if (i==-1)
      ds=-snap;
   else
      ds= snap;

   while (pos<0)
   {
      s+=ds;
      pos+=delta;
   }

   while (pos<cur_v->sy)
   {
      sprintf(temp,"%i",s);
      
      i=pos+gdk_string_height(grid_font,temp)/2;
      if (i>=cur_v->sy)
         break;

      if (i>0)
      {
         DrawStr(2,i,grid_font,temp);
      }
      pos+=delta;
      s+=ds;
   }
}


static void DrawGrids(void)
{
   if (!cur_v->vp.axis_aligned)
      return;
   if (cv_edit_snap_size->i<=0)
      return;
      
   SetColor(C_GRID_1);
   DrawGrid(cv_edit_snap_size->i);
   SetColor(C_GRID_2);
   DrawGrid(cv_edit_snap_size->i*4);
   SetColor(C_GRID_TEXT);
   DrawGridNum();
}


static void ClipDrawLine(vec3_t v1,vec3_t v2)
{
   float   x0, y0, x1, y1;
   float   clipx, clipy;
   int     xs0, ys0;
   int     xs1, ys1;
   int     xs2, ys2;
   int     c0, c1;

   switch (cur_v->mode)
   {
      case M_2D_WIRE:
         if (cur_v->depth_clip)
         {
            if (((v1.z < 0) && (v2.z < 0)) ||
                ((v1.z > cv_depth_clip->f) &&
                 (v2.z > cv_depth_clip->f)) )
            return;
         }

         clipx = dx / cur_v->vp.zoom_amt;
         clipy = dy / cur_v->vp.zoom_amt;

         if (((v1.x > clipx) && (v2.x > clipx)) ||
             ((v1.x <-clipx) && (v2.x <-clipx)) ||
             ((v1.y > clipy) && (v2.y > clipy)) ||
             ((v1.y <-clipy) && (v2.y <-clipy)))
            return;

         x0 =  v1.x * cur_v->vp.zoom_amt + dx;
         y0 = -v1.y * cur_v->vp.zoom_amt + dy;
         x1 =  v2.x * cur_v->vp.zoom_amt + dx;
         y1 = -v2.y * cur_v->vp.zoom_amt + dy;

         xs0 = x0+0.5;
         ys0 = y0+0.5;
         xs1 = x1+0.5;
         ys1 = y1+0.5;

         while (1)
         {
            c0 = OutCode2D(xs0, ys0);
            c1 = OutCode2D(xs1, ys1);

            if ((c0 | c1) == 0)
            {
/*                if (cur_v->vp.grid_type == ALIGN)  TODO: make this work
                  Draw2DSpecialLine(vport,xs0, ys0, xs1, ys1, base_color);
               else*/
                 DrawLine(xs0, ys0, xs1, ys1);
               return;
            }
            else
            if ((c0 & c1) != 0) return;
            else
            if (c0 != 0)
            {
               Clip2D(xs0, ys0, xs1, ys1, &xs2, &ys2, c0);
               xs0 = xs2;
               ys0 = ys2;
            }
            else
            {
               Clip2D(xs0, ys0, xs1, ys1, &xs2, &ys2, c1);
               xs1 = xs2;
               ys1 = ys2;
            }
         }
         break;

      case M_3D_WIRE:
         /* Immediately throw out lines outside view frustum */
         if (OutsideFrustrum(v1,v2)) return;

         if (cur_v->depth_clip)
            if ((v1.z > 1) && (v2.z > 1)) return;

         ClipDrawLine3D(
                     v1.x,v1.y,v1.z,
                     v2.x,v2.y,v2.z);
         break;

   }
}


/* Calculate a svec_t from a vec3_t. */
static int ToScreen(svec_t *d,vec3_t s)
{
   switch (cur_v->mode)
   {
      case M_2D_WIRE:
         if (cur_v->depth_clip)
         {
            if ((s.z < 0) || (s.z > cv_depth_clip->f))
            {
               d->onscreen=0;
               return 0;
            }
         }
         d->x= s.x * cur_v->vp.zoom_amt + dx;
         d->y=-s.y * cur_v->vp.zoom_amt + dy;
         break;

      case M_3D_WIRE:
         if ((s.z <= 0) ||
             (s.x > s.z) ||
             (s.x <-s.z) ||
             (s.y > s.z) ||
             (s.y <-s.z))
         {
            d->onscreen=0;
            return 0;
         }
         if ((cur_v->depth_clip) && (s.z > 1))
         {
            d->onscreen=0;
            return 0;
         }

         d->x = dx * (s.x / s.z + 1);
         d->y = dy * (s.y /-s.z + 1);
         break;

      default:
         d->onscreen=0;
         return 0;
   }
   d->onscreen=1;
   return 1;
}


/* Brush drawing */
static void DrawBrushEdges(brush_t *b)
{
   vec3_t *tverts;
   int from,to;
   int i;

   tverts = b->r->tverts;
   for (i=0;i<b->num_edges;i++)
   {
      from = b->edges[i].startvertex;
      to = b->edges[i].endvertex;
      ClipDrawLine(tverts[from],tverts[to]);
   }
}


static void DrawBrush(brush_t *b)
{
   if (b->bt->flags&BRUSH_F_NORMAL)
   {
      DrawBrushEdges(b);
      return;
   }
/* TODO  if (b->bt->type==BRUSH_Q3_CURVE)
   {
      DrawBrush_Q3Curve(b,color,selected);
      return;
   }*/
/* We don't know how to draw this brush, so we ignore it */
}


static void DrawBrushFace(brush_t *b,int num)
{
   int i;
   face_t *f;
   vec3_t *tverts;
   
   if (!(b->bt->flags&BRUSH_F_NORMAL)) return;

   f=&b->faces[num];
   tverts=b->r->tverts;
   for (i=0;i<f->num_verts-1;i++)
   {
      ClipDrawLine(tverts[f->verts[i]],tverts[f->verts[i+1]]);
   }
   ClipDrawLine(tverts[f->verts[i]],tverts[f->verts[0]]);
}


static void DrawModel(entity_t *e,int selected)
{
   vec3_t trans;
   svec_t scent;
   vec3_t tbox[8];
   vec3_t v;
   int size;

   SetColor(selected?C_ENTITY_SELECT:C_GRID_TEXT); /* TODO */
   v.x=e->min.x; v.y=e->min.y; v.z=e->min.z; Transform(tbox[0],v,cur_v->m);
   v.x=e->max.x;                             Transform(tbox[1],v,cur_v->m);
                 v.y=e->max.y;               Transform(tbox[2],v,cur_v->m);
   v.x=e->min.x;                             Transform(tbox[3],v,cur_v->m);
                 v.y=e->min.y; v.z=e->max.z; Transform(tbox[4],v,cur_v->m);
   v.x=e->max.x;                             Transform(tbox[5],v,cur_v->m);
                 v.y=e->max.y;               Transform(tbox[6],v,cur_v->m);
   v.x=e->min.x;                             Transform(tbox[7],v,cur_v->m);

#define DL(i,j) ClipDrawLine(tbox[i],tbox[j]);
   DL(0,1)
   DL(1,2)
   DL(2,3)
   DL(3,0)
   DL(4,5)
   DL(5,6)
   DL(6,7)
   DL(7,4)
   DL(0,4)
   DL(1,5)
   DL(2,6)
   DL(3,7)
#undef DL

   if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
   {
      Transform(trans,e->center,cur_v->m);
      ToScreen(&scent,trans);

      SetColor(C_ENTITY_FOCAL);

      if (cur_v->mode==M_2D_WIRE)
      {
         size=1.5*cur_v->vp.zoom_amt;
      }
      else
      {
         if (trans.z<0.001)
            size=1/0.001;
         else
            size = 1 / (float)trans.z;
      }
      if (size<1) size=1;
      DrawBox(scent,size);
   }
}


static void DrawViewport(matrix_t m)
{
/* TODO: optimize all this? could probably get rid of tverts and sverts
structures if we transformed stuff while rendering instead of doing it in
two passes.
*/
   int         i, j;
   int         axis_x,axis_y;
   vsel_t     *vs;
   brush_t    *b;
   bsel_t     *bsel;
   fsel_t     *fsel;

   entity_t *e;
   esel_t   *esel;

   face_t *f;

   vec3_t trans;

   int rf=cv_render_flags->i;


   if (rf&(RENDER_BRUSH|RENDER_FACE_FOCAL))
   {
      for (b=M.VBrushHead; b; b=b->VNext)
      {
         b->r->drawn=1;
         if (rf&RENDER_BRUSH)
         {
            for (i=0;i<b->num_verts;i++)
            {
               /* Multiply vertices by transformation matrix*/
               Transform(b->r->tverts[i],b->verts[i],m);
               ToScreen(&b->r->sverts[i],b->r->tverts[i]);
            }
            if (rf&RENDER_BRUSH_FOCAL)
            {
               Transform(trans,b->center,m);
               ToScreen(&b->r->scenter,trans);
            }
         }
         if (rf&RENDER_FACE_FOCAL)
         {
            for (i=0; i<b->num_faces; i++)
            {
               Transform(trans,b->faces[i].center,m);
               ToScreen(&b->faces[i].r->scenter,trans);
            }
         }
      }
      
      if (rf&RENDER_BRUSH_FOCAL)
      {
         for (bsel=M.sel.bsel;bsel;bsel=bsel->Next)
            bsel->Brush->r->drawn=2;
      }
   }

   if (rf&RENDER_ENTITY)
   {
      for (e=M.VEntityHead; e; e=e->VNext)
         e->r->drawn=1;

      if (rf&RENDER_ENTITY_FOCAL)
      {
         for (esel=M.sel.esel;esel;esel=esel->Next)
            esel->Entity->r->drawn=2;
      }
   }


   switch (cur_v->mode)
   {
      case M_2D_WIRE:
         if (cur_v->grid_type == GRID_NORMAL)
            DrawGrids();

      case M_3D_WIRE:
      
         if (rf&RENDER_BRUSH)
         {
            for (b=M.VBrushHead; b; b=b->VNext)
            {
               if (b->r->drawn!=2) /* if ==2, we'll draw it highlighted later anyway */
               {
                  SetColor(GroupColor(UID_Group(b->Group)?UID_Group(b->Group)->color:0));
                  DrawBrush(b);
               }
               if (rf&RENDER_BRUSH_FOCAL)
               {
                  switch (b->bt->type)
                  {
                  case BRUSH_SUBTRACT:
                     SetColor(C_BRUSH_FOCAL_SUB);
                     break;
                  default:
                     SetColor(C_BRUSH_FOCAL);
                     break;
                  }
                  DrawDot(b->r->scenter);
               }
            }
            /* render later so they're on top of the unselected brushes */
            SetColor(C_BRUSH_SELECT);
            for (bsel=M.sel.bsel; bsel; bsel=bsel->Next)
               DrawBrush(bsel->Brush);

            if (rf&RENDER_BRUSH_VERT)
            {
               SetColor(C_VERT);
               for (bsel=M.sel.bsel; bsel; bsel=bsel->Next)
                  for (j=0; j<bsel->Brush->num_verts; j++)
                     DrawVertex(bsel->Brush->r->sverts[j]);
            }
         }

         if (rf&RENDER_FACE_FOCAL)
         {
            SetColor(C_FACE_SELECT);
            for (fsel=M.sel.fsel; fsel; fsel=fsel->Next)
               DrawBrushFace(fsel->Brush, fsel->num);

            SetColor(C_FACE_FOCAL);
            for (b=M.VBrushHead; b; b=b->VNext)
            {
               if (b->bt->type==BRUSH_Q3_CURVE) continue;

               for (i=0; i<b->num_faces; i++)
                  DrawDot(b->faces[i].r->scenter);
            }

            SetColor(C_FACE_SELECT);
            for (fsel=M.sel.fsel; fsel; fsel=fsel->Next)
               DrawDot(fsel->Brush->faces[fsel->num].r->scenter);

            if (rf&RENDER_FACE_VERT)
            {
               SetColor(C_VERT);
               for (fsel=M.sel.fsel; fsel; fsel=fsel->Next)
               {
                  b = fsel->Brush;
                  f = &fsel->Brush->faces[fsel->num];
                  for (j=0; j<f->num_verts; j++)
                     DrawVertex(b->r->sverts[f->verts[j]]);
               }
            }
         }

         if (rf&RENDER_ENTITY)
         {
            for (e=M.VEntityHead; e; e=e->VNext)
            {
               if (e->r->drawn==2) /* highlighted later if ==2 */
                  continue;
               DrawModel(e,0);
            }

            if (rf&RENDER_ENTITY_FOCAL)
            {
               for (esel=M.sel.esel; esel; esel=esel->Next)
               {
                  e = esel->Entity;
                  DrawModel(e,1);
               }
            }

            if (rf&RENDER_ENTITY_LINK)
/* TODO              DrawAllLinks()*/;
         }

         if (rf&(RENDER_BRUSH_VERT|RENDER_FACE_VERT))
         {
            SetColor(C_VERT_SEL);
            for (vs=M.sel.vsel; vs; vs=vs->Next)
               DrawVertex(vs->Brush->r->sverts[vs->num]);
         }
               
         break;


#if TODO
      case SOLID:
         DrawPolyView(cur_vport);
         switch (status.edit_mode)
         {
            case BRUSH:
               DBrushSel();

               DBrushSelVer();

               DVerSel();

               break;
               

            case FACE:
               /* Highlight all selected faces */
               for (fsel=M.display.fsel; fsel; fsel=fsel->Next)
               {
                  DrawBrushFace(fsel->Brush, fsel->facenum, CR_FACE_SELECT, TRUE);
               }
               /* Draw vertices of selected faces */
               for (fsel=M.display.fsel; fsel; fsel=fsel->Next)
               {
                  b = fsel->Brush;
                  p = &fsel->Brush->faces[fsel->facenum];
                  for (j=0; j<p->num_verts; j++)
                  {
                     DrawVertex(cur_vport, b->sverts[p->verts[j]], CR_VERT);
                  }
               }

               DVerSel();

               break;
               

            case MODEL:
               DBrushSel();
               break;
         }
         break;
#endif
   }

/* Draw axises */
   {
      vec3_t   a1,a2;
      int      i;
      scalar_t ax,ay;
      matrix_t m;
      float    rx,ry,rz;
      GdkColor *col;

      axis_x = 20;
      axis_y = cur_v->sy - 20;

      col=0;

      GetRotValues(&cur_v->vp,&rx,&ry,&rz);
      GenerateRotMatrix(m,rx,ry,rz);

      for (i=0;i<3;i++)
      {
         a1.x=a1.y=a1.z=0;
         switch (i)
         {
         case 0:
            a1.x=1;
            break;
         case 1:
            a1.y=1;
            break;
         case 2:
            a1.z=1;
            break;
         }

         Transform(a2,a1,m);

         ax= a2.x*15;
         ay=-a2.y*15;

         switch (i)
         {
         case 0:
            col=C_AXIS_X;
            break;
         case 1:
            col=C_AXIS_Y;
            break;
         case 2:
            col=C_AXIS_Z;
            break;
         }

         SetColor(col);
         DrawLine(axis_x,axis_y,axis_x+ax,axis_y+ay);
      }
   }

/*   DrawPts(m);  TODO*/
}

   
void S_3D_Render(g_viewport_t *v)
{
   if (!v->pic) return;

   /* Start by clearing viewport */
   gdk_gc_set_foreground(gc,C_BACK);
   cur_color=C_BACK;
   gdk_draw_rectangle(
      v->pic,
      gc,
      TRUE,
      0,0,v->sx,v->sy);

   /* If the viewport_t values aren't valid yet, don't try to render it. */
   if (!v->vp.active) return;
   
   cur_v=v;
   dx=cur_v->sx/2;
   dy=cur_v->sy/2;
   
   GenerateMatrix(v->m,v);
   
   DrawViewport(v->m);
}


void S_3D_Init(g_viewport_t *v)
{
   int i;
   GdkColor *c;

   cv_depth_clip=Cfg_DefVar("gtk.3d.depth_clip","800",
      "The distance beyond which things are clipped if depth clipping is enabled in a viewport.");
   cv_fov=Cfg_DefVar("gtk.3d.fov","80",
      "The horizontal field of view angle.");
   cv_grid_font=Cfg_DefVar("gtk.3d.grid_font","-*-clean-medium-r-*-*-*-120-*-*-*-*-*-*",
      "The font used for the numbers in the grid.");

   gc=gdk_gc_new(GTK_WIDGET(v->win)->window);

   grid_font=gdk_font_load(cv_grid_font->s);
   if (!grid_font) S_Abort("S_3D_Init","Can't load font for grid!");

   for (i=0,c=colors;i<C_MAX;i++,c++)
      gdk_colormap_alloc_color(S_colormap,c,FALSE,TRUE);
}


int S_R_Trans_Vert(vec3_t v,vec3_t *sv)
{
   svec_t s;
   vec3_t t;

   Transform(t,v,S_cur_gvport->m);
   cur_v=S_cur_gvport;
   dx=cur_v->sx/2;
   dy=cur_v->sy/2;
   if (!ToScreen(&s,t)) return 0;
   sv->x=s.x;
   sv->y=s.y;
   sv->z=t.z; /* good measure of distance from screen for sorting selection lists */
   return 1;
}

int S_R_Trans_SVert(vec3_t sv,vec3_t *v)
{
   int dx,dy;
   vec3_t t;
   matrix_t m;
   
   if (S_cur_gvport->mode!=M_2D_WIRE)
      return 0;

   GenerateIMatrix(m,S_cur_gvport);

   dx=S_cur_gvport->sx>>1;
   dy=S_cur_gvport->sy>>1;

   t.x= (sv.x - dx)/S_cur_gvport->vp.zoom_amt;
   t.y=-(sv.y - dy)/S_cur_gvport->vp.zoom_amt;
   t.z=1;

   v->x=m[0][0]*t.x+
        m[0][1]*t.y+
        m[0][2]*t.z+
        m[0][3];
   v->y=m[1][0]*t.x+
        m[1][1]*t.y+
        m[1][2]*t.z+
        m[1][3];
   v->z=m[2][0]*t.x+
        m[2][1]*t.y+
        m[2][2]*t.z+
        m[2][3];

   return 1;
}

int S_R_GetVector(vec3_t spos,vec3_t *org,vec3_t *dir)
{
   matrix_t m;
   vec3_t s;

   /* TODO: implement this for M_2D_WIRE too */
   if (S_cur_gvport->mode!=M_3D_WIRE)
      return 0;

   GenerateIMatrix(m,S_cur_gvport);

   s.x= (spos.x)*2.0/(S_cur_gvport->sx)-1;
   s.y=-(spos.y)*2.0/(S_cur_gvport->sy)+1;
   s.z=1;

   dir->x=s.x*m[0][0]+
          s.y*m[0][1]+
          s.z*m[0][2];
   dir->y=s.x*m[1][0]+
          s.y*m[1][1]+
          s.z*m[1][2];
   dir->z=s.x*m[2][0]+
          s.y*m[2][1]+
          s.z*m[2][2];

   Normalize(dir);

   org->x=S_cur_gvport->vp.camera_pos.x;
   org->y=S_cur_gvport->vp.camera_pos.y;
   org->z=S_cur_gvport->vp.camera_pos.z;

   return 1;
}

int S_R_GetMousePos(vec3_t *spos)
{
   int x,y;

   gdk_window_get_pointer(GTK_WIDGET(S_cur_gvport->win)->window,&x,&y,NULL);
   if (x<0 || y<0 || x>=S_cur_gvport->sx || y>=S_cur_gvport->sy)
      return 0;
   spos->x=x;
   spos->y=y;
   spos->z=0;
   return 1;
}

/* TODO: make this configurable */
#define NEAR 8
int S_R_IsNear(vec3_t sv1,vec3_t sv2)
{
   if (fabs(sv1.x-sv2.x)<NEAR &&
       fabs(sv1.y-sv2.y)<NEAR)
      return 1;
   return 0;
}

int S_R_IsInBox(vec3_t sv,vec3_t min,vec3_t max)
{
   if (min.x>sv.x || min.y>sv.y) return 0;
   if (max.x<sv.x || max.y<sv.y) return 0;
   return 1;
}


/* Used in g_int.c */
void S_SelectionWindow_DrawBox(int x0,int y0,int x1,int y1)
{
   int sx,sy,w,h;

   if (x0<x1)
      sx=x0,w=x1-x0;
   else
      sx=x1,w=x0-x1;
         
   if (y0<y1)
      sy=y0,h=y1-y0;
   else
      sy=y1,h=y0-y1;
      
   SetColor(C_SELECTION_BOX);
   gdk_draw_rectangle(GTK_WIDGET(S_cur_gvport->da)->window,
      gc,0,
      sx,sy,w,h);
}

