/*
newgroup.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "group.h"

#include "error.h"
#include "memory.h"
#include "quest.h"
#include "uid.h"
#include "undo.h"
#include "fe/message.h"


/* TODO: handle group colors properly
have a MAX_COLORS thing, color starts at 0 and can be cycled through MAX_COLORS-1.
System dependent stuff keeps track of converting color number to real color, define
somewhere what colors the different numbers 'should' be.
*/


group_t *G_New(void)
{
   group_t *g;

	g=Mem_Alloc(&mem_group);
   if (!g) return NULL;
   memset(g,0,sizeof(group_t));
   return g;
}

void G_FreeC(group_t *g)
{ /* Nothing to do, no fields are allocated */
   ASSERT(!g->real || (!g->Next && !g->Last));
}

void G_Free(group_t *g)
{
   G_FreeC(g);
	Mem_Free(&mem_group,g);
}

void G_Link(group_t *g)
{
   ASSERT(!g->Next && !g->Last);

   if (M.GroupHead!=NULL) M.GroupHead->Last = g;
   g->Next = M.GroupHead;
   g->Last = NULL;
   M.GroupHead = g;
   g->real=1;

   g->uid=GetGroupUID();

   M.num_groups++;

   U_GroupAdd(g);
}

void G_Unlink(group_t *g)
{
   ASSERT(!(g->parent || g->child || g->cnext || g->clast));
   ASSERT(g->Next || g->Last || g->uid==M.WorldGroup);
   ASSERT(g->real);

   U_GroupDelete(g);

   M.num_groups--;
   
   if (g->Next != NULL) g->Next->Last = g->Last;
   if (g->Last != NULL) g->Last->Next = g->Next;
   if (g==M.GroupHead) M.GroupHead=g->Next;

   g->Next=NULL;
   g->Last=NULL;
   g->real=0;
}

group_t *G_Duplicate(group_t *dest,group_t *src)
{
   if (!dest)
   {
		dest=Mem_Alloc(&mem_group);
      if (!dest) return NULL;
   }
   memset(dest,0,sizeof(group_t));

   strcpy(dest->groupname,src->groupname);
   dest->color=src->color;
   dest->flags=src->flags;
   dest->parent=src->parent;
   dest->child=src->child;
   dest->cnext=src->cnext;
   dest->clast=src->clast;

   return dest;
}


int G_Size(ARG_UNUSED(group_t *g))
{
   return 0;
}

void G_Update(ARG_UNUSED(group_t *g))
{
}


void Grp_CreateWorldGroup(void)
{
   M.GroupHead=G_New();
   strcpy(M.GroupHead->groupname,"World");
   strcpy(M.GroupHead->fullname,"World");
   M.GroupHead->flags=GROUP_WORLD;
   M.num_groups=1;
   M.CurGroup=M.WorldGroup=M.GroupHead->uid=GetGroupUID();
   M.GroupHead->real = 1;
}


map_uid_t Grp_Find(const char *fullname)
{
   group_t *g;

   for (g=M.GroupHead;g;g=g->Next)
      if (!stricmp(fullname,g->fullname))
         return g->uid;
   return 0;
}


map_uid_t Grp_Create(const char *groupname,map_uid_t uparent)
{
	group_t *g;
   group_t *parent;
   map_uid_t ug;

   if (uparent)
      parent=UID_Group(uparent);
   else
      parent=UID_Group(M.WorldGroup);

	for (ug=parent->child;ug;ug=g->cnext)
   {
      g=UID_Group(ug);
		if (stricmp(g->groupname, groupname) == 0)
			return 0;
   }

   g=G_New();
   strcpy(g->groupname,groupname);
   if (parent->uid!=M.WorldGroup)
   {
      strcpy(g->fullname,parent->fullname);
      strcat(g->fullname,".");
   }
   else
      g->fullname[0]=0;
   strcat(g->fullname,groupname);
   g->hidden=parent->hidden;

   G_Link(g);

   g->parent=uparent;
   g->cnext=parent->child;
   g->clast=0;
   if (parent->child)
   {
      group_t *g2=UID_Group(parent->child);
      U_GroupModify(g2);
      g2->clast=g->uid;
   }
   U_GroupModify(parent);
   parent->child=g->uid;
   

	return g->uid;
}


void Grp_Delete(map_uid_t grp)
{
	group_t *g,*cn,*cl,*p;

/* TODO: return error value instead */
   if (grp==M.WorldGroup)
   {
      ERROR("Group delete: Deleting world group not allowed!");
      return;
   }
   g=UID_Group(grp);
   if (!g)
   {
      ERROR("Group delete: Can't find group!");
      return;
   }

   if (g->child)
   {
      ERROR("Group delete: Deleting group with children not allowed!");
      return;
   }

   cn=UID_Group(g->cnext);
   cl=UID_Group(g->clast);

   if (cn)
   {
      U_GroupModify(cn);
      cn->clast=g->clast;
   }
   if (cl)
   {
      U_GroupModify(cl);
      cl->cnext=g->cnext;
   }
   else
   {
      p=UID_Group(g->parent);
      U_GroupModify(p);
      p->child=g->cnext;
   }

   g->cnext=g->clast=g->parent=0;

   G_Unlink(g);
   G_Free(g);
}


static void update_r(group_t *g,int hidden)
{
   map_uid_t uc;
   group_t *c;
   
   if (g->flags&GROUP_HIDDEN)
      hidden=1;
   g->hidden=hidden;

   for (uc=g->child;uc;)
   {
      c=UID_Group(uc);
      strcpy(c->fullname,g->fullname);
      if (c->fullname[0])
         strcat(c->fullname,".");
      strcat(c->fullname,c->groupname);
      update_r(c,hidden);
      uc=c->cnext;
   }
}

void Grp_Update(void)
{
   group_t *g=UID_Group(M.WorldGroup);
	if (!g)
	{
		WARNING("Can't find world group!");
		return;
	}
   g->fullname[0]=0;
   update_r(g,0);
   strcpy(g->fullname,g->groupname);
}


map_uid_t FindVisGroup(map_uid_t first)
{
   group_t *g;

   if (first && (g=UID_Group(first)))
   {
      if (!(g->hidden))
         return first;
   }
   return GetDefaultGroup();
}

map_uid_t GetDefaultGroup(void)
{
   group_t *g;
   map_uid_t ug;

   for (ug=M.CurGroup;ug;ug=g->parent)
   {
      g=UID_Group(ug);
      if (g && !(g->hidden))
         return ug;
   }
   if (M.WorldGroup)
      return M.WorldGroup;
   Warning("GetDefaultGroup: No group found, world group gone! (internal error?)");
   return 0;
}

